# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import errno
import glob
import logging
import os
import os.path
from csv import DictReader
from io import StringIO

import gevent
from gevent.lock import Semaphore

from volttron.client.known_identities import CONFIGURATION_STORE
from volttron.client.vip.agent import RPC, Agent, Core, Unreachable, VIPError
from volttron.server.decorators import service
from volttron.server.server_options import ServerOptions
from volttron.types import Service, Identity
from volttron.types.agent_context import AgentOptions
from volttron.types.auth import Credentials, CredentialsStore
from volttron.types.service_interface import ServiceInterface
from volttron.utils import (format_timestamp, get_aware_utc_now, jsonapi, parse_json_config)
from volttron.utils.jsonrpc import MethodNotFound, RemoteError
from volttron.utils.persistance import PersistentDict
from volttron.utils.storeutils import (check_for_recursion, store_ext, strip_config_name)

_log = logging.getLogger(__name__)

UPDATE_TIMEOUT = 30.0


def process_store(identity, store):
    """Parses raw store data and returns contents.
    Called at startup to initialize the parsed version of the store."""
    results = {}
    name_map = {}
    sync_store = False
    for config_name, config_data in store.items():
        config_type = config_data["type"]
        config_string = config_data["data"]
        try:
            processed_config = process_raw_config(config_string, config_type)
            if check_for_recursion(config_name, processed_config, results):
                raise ValueError("Recursive configuration references")
            results[config_name] = processed_config
        except ValueError as e:
            _log.error("Error processing Agent {} config {}: {}".format(
                identity, config_name, str(e)))
            sync_store = True
            del store[config_name]

        if config_name.lower() in name_map:
            _log.error("Conflicting names in store, dropping {}".format(config_name))
            sync_store = True
            del store[config_name]

        else:
            name_map[config_name.lower()] = config_name

    if sync_store:
        _log.warning("Removing invalid configurations for Agent {}".format(identity))
        store.sync()

    return results, name_map


def process_raw_config(config_string, config_type="raw"):
    """Parses raw config string into python objects"""
    if config_type == "raw":
        return config_string
    elif config_type == "json":
        config = parse_json_config(config_string)
        if not isinstance(config, (list, dict)):
            raise ValueError("Configuration must be a list or object.")
        return config
    elif config_type == "csv":
        f = StringIO(config_string)
        return [x for x in DictReader(f)]

    raise ValueError("Unsupported configuration type.")


@service
class ConfigStoreService(Agent):

    class Meta:
        identity = CONFIGURATION_STORE

    def __init__(self,
                 options: ServerOptions,
                 credential_store: CredentialsStore | None = None,
                 **kwargs):
        # kwargs["enable_store"] = False
        # kwargs["identity"] = self.Meta.identity
        agent_options = AgentOptions(enable_store=False)

        if credential_store is not None:
            creds = credential_store.retrieve_credentials(identity=self.Meta.identity)
        else:
            creds = Credentials(identity=self.Meta.identity)

        super().__init__(credentials=creds,
                         address=options.service_address,
                         options=agent_options,
                         **kwargs)

        # This agent is started before the router so we need
        # to keep it from blocking.
        self.core.delay_running_event_set = False

        self.store = {}
        self.store_path = os.path.join(os.environ["VOLTTRON_HOME"], "configuration_store")
        # entry = AuthEntry(credentials=self.core.publickey,
        #                   user_id=self.core.identity,
        #                   capabilities="sync_agent_config",
        #                   comments="Automatically added by config store service")
        # AuthFile().add(entry, overwrite=True)

    @Core.receiver("onsetup")
    def _setup(self, sender, **kwargs):
        _log.info("Initializing configuration store service.")

        try:
            # explicitly provide access to others. Needed for secure mode.
            # Agents needs access to its own store file in this dir
            os.makedirs(self.store_path)
            os.chmod(self.store_path, 0o755)
        except OSError as e:
            if e.errno != errno.EEXIST:
                _log.critical("Failed to create configuration store directory: " + str(e))
                raise
            else:
                _log.debug("Configuration directory already exists.")

        config_store_iter = glob.iglob(os.path.join(self.store_path, "*" + store_ext))

        for store_path in config_store_iter:
            root, ext = os.path.splitext(store_path)
            agent_identity = os.path.basename(root)
            _log.debug("Processing store for agent {}".format(agent_identity))
            # store = PersistentDict(filename=store_path, flag="c", format="json")
            store = {}
            if os.path.exists(store_path):
                store = jsonapi.loads(open(store_path).read())

            parsed_configs, name_map = process_store(agent_identity, store)
            self.store[agent_identity] = {
                "configs": parsed_configs,
                "store": store,
                "name_map": name_map,
                "lock": Semaphore(),
            }

    @RPC.export
    def set_config(self,
                   identity,
                   config_name,
                   raw_contents,
                   config_type="raw",
                   trigger_callback=True,
                   send_update=True):
        contents = process_raw_config(raw_contents, config_type)
        self._add_config_to_store(identity,
                                  config_name,
                                  raw_contents,
                                  contents,
                                  config_type,
                                  trigger_callback=trigger_callback,
                                  send_update=send_update)

    @RPC.export
    def delete_config(self, identity, config_name, trigger_callback=True, send_update=True):
        self.delete(identity,
                    config_name,
                    trigger_callback=trigger_callback,
                    send_update=send_update)

    @RPC.export
    def delete_store(self, identity):
        agent_store = self.store.get(identity)
        if agent_store is None:
            return

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_store_lock = agent_store["lock"]
        agent_name_map = agent_store["name_map"]

        agent_configs.clear()
        agent_disk_store.clear()
        agent_name_map.clear()

        # Sync will delete the file if the store is empty.
        # agent_disk_store.async_sync()
        if not agent_disk_store and os.path.exists(
                os.path.join(self.store_path, identity + store_ext)):
            os.remove(os.path.join(self.store_path, identity + store_ext))

        if identity in self.vip.peerlist.peers_list:
            with agent_store_lock:
                try:
                    self.vip.rpc.call(identity,
                                      "config.update",
                                      "DELETE_ALL",
                                      None,
                                      trigger_callback=True).get(timeout=UPDATE_TIMEOUT)
                except Unreachable:
                    _log.debug(
                        "Agent {} not currently running. Configuration update not sent.".format(
                            identity))
                except RemoteError as e:
                    _log.error("Agent {} failure when all configurations: {}".format(identity, e))
                except MethodNotFound as e:
                    _log.error("Agent {} failure when deleting configuration store: {}".format(
                        identity, e))

        # If the store is still empty (nothing jumped in and added to it while
        # we were informing the agent) then remove it from the global store.
        if not agent_disk_store:
            self.store.pop(identity, None)

    @RPC.export
    def list_configs(self, identity):
        result = list(self.store.get(identity, {}).get("store", {}).keys())
        result.sort()
        return result

    @RPC.export
    def list_stores(self):
        result = list(self.store.keys())
        result.sort()
        return result

    @RPC.export
    def get_config(self, identity, config_name, raw=True):
        agent_store = self.store.get(identity)
        if agent_store is None:
            raise KeyError('No configuration file "{}" for VIP IDENTITY {}'.format(
                config_name, identity))

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_name_map = agent_store["name_map"]

        config_name = strip_config_name(config_name)
        config_name_lower = config_name.lower()

        if config_name_lower not in agent_name_map:
            raise KeyError('No configuration file "{}" for VIP IDENTITY {}'.format(
                config_name, identity))

        real_config_name = agent_name_map[config_name_lower]

        if raw:
            return agent_disk_store[real_config_name]["data"]

        return agent_configs[real_config_name]

    @RPC.export
    def get_metadata(self, identity, config_name):
        agent_store = self.store.get(identity)
        if agent_store is None:
            raise KeyError('No configuration file "{}" for VIP IDENTITY {}'.format(
                config_name, identity))

        agent_disk_store = agent_store["store"]
        agent_name_map = agent_store["name_map"]

        config_name = strip_config_name(config_name)
        config_name_lower = config_name.lower()

        if config_name_lower not in agent_name_map:
            raise KeyError('No configuration file "{}" for VIP IDENTITY {}'.format(
                config_name, identity))

        real_config_name = agent_name_map[config_name_lower]

        real_config = agent_disk_store[real_config_name]

        # Set modified to none if we predate the modified flag.
        if real_config.get("modified") is None:
            real_config["modified"] = None

        return real_config

    @RPC.export
    def initialize_configs(self, identity: Identity):
        """
        Called by an Agent at startup to trigger initial configuration state
        push.
        """
        _log.debug("Initializing configurations for agent {}".format(identity))
        # We need to create store and lock if it doesn't exist in case someone
        # tries to add a configuration while we are sending the initial state.
        agent_store = self.store.get(identity)

        if agent_store is None:
            # Initialize a new store.
            store_path = os.path.join(self.store_path, identity + store_ext)
            store = PersistentDict(filename=store_path, flag="c", format="json")
            agent_store = {
                "configs": {},
                "store": store,
                "name_map": {},
                "lock": Semaphore(),
            }
            self.store[identity] = agent_store

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_store_lock = agent_store["lock"]
        if identity in self.vip.peerlist.peers_list:
            with agent_store_lock:
                try:
                    self.vip.rpc.call(identity, "config.initial_update",
                                      agent_configs).get(timeout=UPDATE_TIMEOUT)
                except Unreachable:
                    _log.debug(
                        "Agent {} not currently running. Configuration update not sent.".format(
                            identity))
                except RemoteError as e:
                    _log.error("Agent {} failure when performing initial update: {}".format(
                        identity, e))
                except MethodNotFound as e:
                    _log.error("Agent {} failure when performing initial update: {}".format(
                        identity, e))
                except VIPError as e:
                    _log.error("VIP Error sending initial agent configuration: {}".format(e))

        # If the store is empty (and nothing jumped in and added to it while we
        # were informing the agent) then remove it from the global store.
        if not agent_disk_store:
            self.store.pop(identity, None)

    # Helper method to allow the local services to delete configs before message
    # bus in online.
    def delete(self, identity, config_name, trigger_callback=False, send_update=True):
        agent_store = self.store.get(identity)
        if agent_store is None:
            raise KeyError('No configuration file "{}" for VIP IDENTITY {}'.format(
                config_name, identity))

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_store_lock = agent_store["lock"]
        agent_name_map = agent_store["name_map"]

        config_name = strip_config_name(config_name)
        config_name_lower = config_name.lower()

        if config_name_lower not in agent_name_map:
            raise KeyError('No configuration file "{}" for VIP IDENTITY {}'.format(
                config_name, identity))

        real_config_name = agent_name_map[config_name_lower]

        agent_configs.pop(real_config_name)
        agent_disk_store.pop(real_config_name)
        agent_name_map.pop(config_name_lower)

        # Sync will delete the file if the store is empty.
        if not agent_disk_store and os.path.exists(
                os.path.join(self.store_path, identity + store_ext)):
            os.remove(os.path.join(self.store_path, identity + store_ext))
        else:
            with open(os.path.join(self.store_path, identity + store_ext), "w") as f:
                jsonapi.dump(agent_disk_store, f)

        if send_update and identity in self.vip.peerlist.peers_list:
            with agent_store_lock:
                try:
                    self.vip.rpc.call(
                        identity,
                        "config.update",
                        "DELETE",
                        config_name,
                        trigger_callback=trigger_callback,
                    ).get(timeout=UPDATE_TIMEOUT)
                except Unreachable:
                    _log.debug(
                        "Agent {} not currently running. Configuration update not sent.".format(
                            identity))
                except RemoteError as e:
                    _log.error("Agent {} failure when deleting configuration {}: {}".format(
                        identity, config_name, e))
                except MethodNotFound as e:
                    _log.error("Agent {} failure when adding/updating configuration {}: {}".format(
                        identity, config_name, e))

        # If the store is empty (and nothing jumped in and added to it while we
        # were informing the agent) then remove it from the global store.
        if not agent_disk_store:
            self.store.pop(identity, None)

    # Helper method to allow the local services to store configs before message
    # bus is online.
    def store_config(self,
                     identity,
                     config_name,
                     contents,
                     trigger_callback=False,
                     send_update=True):
        config_type = None
        raw_data = None
        if isinstance(contents, (dict, list)):
            config_type = "json"
            raw_data = jsonapi.dumps(contents)
        elif isinstance(contents, str):
            config_type = "raw"
            raw_data = contents
        else:
            raise ValueError("Unsupported configuration content type: {}".format(
                str(type(contents))))

        self._add_config_to_store(identity,
                                  config_name,
                                  raw_data,
                                  contents,
                                  config_type,
                                  trigger_callback=trigger_callback,
                                  send_update=send_update)

    def _add_config_to_store(self,
                             identity,
                             config_name,
                             raw,
                             parsed,
                             config_type,
                             trigger_callback=False,
                             send_update=True):
        """Adds a processed configuration to the store."""
        agent_store = self.store.get(identity)

        action = "UPDATE"

        if agent_store is None:
            # Initialize a new store.
            store_path = os.path.join(self.store_path, identity + store_ext)
            store = PersistentDict(filename=store_path, flag="c", format="json")
            agent_store = {
                "configs": {},
                "store": store,
                "name_map": {},
                "lock": Semaphore(),
            }
            self.store[identity] = agent_store

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_store_lock = agent_store["lock"]
        agent_name_map = agent_store["name_map"]

        config_name = strip_config_name(config_name)
        config_name_lower = config_name.lower()

        if config_name_lower not in agent_name_map:
            action = "NEW"

        if check_for_recursion(config_name, parsed, agent_configs):
            raise ValueError("Recursive configuration references detected.")

        if config_name_lower in agent_name_map:
            old_config_name = agent_name_map[config_name_lower]
            del agent_configs[old_config_name]

        agent_configs[config_name] = parsed
        agent_name_map[config_name_lower] = config_name

        agent_disk_store[config_name] = {
            "type": config_type,
            "modified": format_timestamp(get_aware_utc_now()),
            "data": raw,
        }

        if not agent_disk_store and os.path.exists(
                os.path.join(self.store_path, identity + store_ext)):
            os.remove(os.path.join(self.store_path, identity + store_ext))
        else:
            with open(os.path.join(self.store_path, identity + store_ext), "w") as f:
                jsonapi.dump(agent_disk_store, f)

        _log.debug("Agent {} config {} stored.".format(identity, config_name))

        if send_update and identity in self.vip.peerlist.peers_list:
            with agent_store_lock:
                try:
                    self.vip.rpc.call(
                        identity,
                        "config.update",
                        action,
                        config_name,
                        contents=parsed,
                        trigger_callback=trigger_callback,
                    ).get(timeout=UPDATE_TIMEOUT)
                except Unreachable:
                    _log.debug(
                        "Agent {} not currently running. Configuration update not sent.".format(
                            identity))
                except RemoteError as e:
                    _log.error("Agent {} failure when adding/updating configuration {}: {}".format(
                        identity, config_name, e))
                except MethodNotFound as e:
                    _log.error("Agent {} failure when adding/updating configuration {}: {}".format(
                        identity, config_name, e))
                except gevent.timeout.Timeout:
                    _log.error("Config update to agent {} timed out after {} seconds".format(
                        identity, UPDATE_TIMEOUT))
                except Exception as e:
                    _log.error("Unknown error sending update to agent identity {}.: {}".format(
                        identity, e))



from __future__ import annotations

import argparse
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Protocol, runtime_checkable

from volttron.types.agent_context import AgentContext
from volttron.types.auth.auth_credentials import Credentials
from volttron.types import Connection, CoreLoop

if TYPE_CHECKING:
    from volttron.client.vip.agent import Agent


class ControlParser(ABC):

    @abstractmethod
    def get_parser(self) -> argparse.Parser:
        ...


class ConnectionBuilder(ABC):

    @abstractmethod
    def build(self, *, credentials: Credentials) -> Connection:
        ...


class CoreBuilder(ABC):

    @abstractmethod
    def build(self, *, context: AgentContext, owner: Agent = None) -> CoreLoop:
        ...

    # def __init__(self, core_cls: type[Core], connection_factory: ConnectionBuilder) -> None:
    #     self._core_cls = core_cls
    #     self._connection_factory = connection_factory

    # def create(self, credentials: Credentials, owner: Agent = None) -> Core:
    #     core = self._core_cls(credentials=credentials, connection_factory=self._connection_factory)
    #     return core

    # def register(cls: )



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
# isort: skip_file
from __future__ import annotations

import logging
import os

from volttron.types.auth.auth_credentials import Credentials
from volttron.types.agent_context import AgentContext, AgentOptions
from volttron.types import AbstractAgent

from .core import *
from .decorators import *
from .errors import *
from .subsystems import *

_log = logging.getLogger(__name__)


class Agent(AbstractAgent):

    class Subsystems(object):

        def __init__(self, *, owner: Agent, core: Core, options: AgentOptions):
            # heartbeat_autostart, heartbeat_period, enable_store, enable_web, enable_channel,
            #         message_bus, tag_vip_id, tag_refresh_interval):
            self.peerlist = PeerList(core=core)
            self.ping = Ping(core)
            self.rpc = RPC(core=core, owner=owner, peerlist_subsys=self.peerlist)
            self.hello = Hello(core=core)
            self.pubsub = PubSub(core=core,
                                 peerlist_subsys=self.peerlist,
                                 rpc_subsys=self.rpc,
                                 owner=self,
                                 tag_vip_id=options.tag_vip_id,
                                 tag_refresh_interval=options.tag_refresh_interval)
            self.health = Health(owner=owner, core=core, rpc=self.rpc)
            self.heartbeat = Heartbeat(owner,
                                       core,
                                       rpc=self.rpc,
                                       pubsub=self.pubsub,
                                       heartbeat_autostart=options.heartbeat_autostart,
                                       heartbeat_period=options.heartbeat_period)
            if options.enable_store:
                self.config = ConfigStore(owner, core, self.rpc)

            self.auth = Auth(owner, core, self.rpc)

    def __init__(self, *, credentials: Credentials = None, options: AgentOptions = None, address: str = None, **kwargs):

        from volttron.client.decorators import (get_connection_builder, get_core_builder)

        factory = get_core_builder(**kwargs)

        if options is None:
            options = AgentOptions()

        if credentials is None:
            if not (identity := os.environ.get('AGENT_VIP_IDENTITY')):
                raise ValueError(f"Environmental variable AGENT_VIP_IDENTITY not set!")

            credentials = self.get_credentials(identity)

        # TODO: We need to be able to get the address from environment probably here.
        context = AgentContext(credentials=credentials, options=options, address=address)

        # Build the core based upon what was loaded in the factory above.
        self.core = factory.build(owner=self, context=context)

        self.vip = Agent.Subsystems(owner=self, core=self.core, options=options)
        self.core.setup()
        self.vip.rpc.export(self.core.version, "agent.version")
        # except Exception as e:
        #     _log.exception("Exception creating Agent. {}".format(e))
        #     raise e


class BasicAgent(object):

    def __init__(self, **kwargs):
        kwargs.pop("identity", None)
        super(BasicAgent, self).__init__(**kwargs)
        self.core = BasicCore(self)


def build_agent(*, address=None, credentials: Credentials):
    raise NotImplementedError()


# def build_agent(address=None,
#                 identity=None,
#                 publickey=None,
#                 secretkey=None,
#                 timeout=10,
#                 serverkey=None,
#                 agent_class=Agent,
#                 volttron_central_address=None,
#                 volttron_central_instance_name=None,
#                 **kwargs) -> Agent:
#     """Builds a dynamic agent connected to the specifiedd address.

#     All key parameters should have been encoded with
#     :py:meth:`volttron.client.vip.socket.encode_key`

#     :param str address: VIP address to connect to
#     :param str identity: Agent's identity
#     :param str publickey: Agent's Base64-encoded CURVE public key
#     :param str secretkey: Agent's Base64-encoded CURVE secret key
#     :param str serverkey: Server's Base64-encoded CURVE public key
#     :param class agent_class: Class to use for creating the instance
#     :param int timeout: Seconds to wait for agent to start
#     :param kwargs: Any Agent specific parameters
#     :return: an agent based upon agent_class that has been started
#     :rtype: agent_class
#     """

#     address = address if address is not None else get_address()

#     # This is a fix allows the connect to message bus to be different than
#     # the one that is currently running.
#     if publickey is None or secretkey is None:
#         publickey, secretkey = get_server_keys()

#     message_bus = cc.get_messagebus()

#     try:
#         enable_store = kwargs.pop("enable_store")
#     except KeyError:
#         enable_store = False

#     agent = agent_class(address=address,
#                         identity=identity,
#                         publickey=publickey,
#                         secretkey=secretkey,
#                         serverkey=serverkey,
#                         volttron_central_address=volttron_central_address,
#                         volttron_central_instance_name=volttron_central_instance_name,
#                         message_bus=message_bus,
#                         enable_store=enable_store,
#                         **kwargs)
#     event = gevent.event.Event()
#     gevent.spawn(agent.core.run, event)
#     with gevent.Timeout(timeout):
#         event.wait()
#     return agent



from volttron_core_fixtures.environment import (
    volttron_home,
    isolated_volttron_home,
    persistent_volttron_home,
    configured_volttron_home,
    create_volttron_home_with_config,
    create_volttron_home,
    create_volttron_home_fun_scope,
    create_volttron_home_mod_scope
)


# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2023 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging
import os
import sys
import collections

from volttron.utils import jsonapi
from volttron.client.known_identities import AUTH
from volttron.types.auth import AuthException

# TODO Look and see what query subsystem is going to work
#from volttron.platform.vip.agent.subsystems.query import Query

_log = logging.getLogger(__name__)

_stdout = sys.stdout
_stderr = sys.stderr


# TODO: Re-evaluate for usage
def list_remotes(opts):
    """Lists remote certs and credentials.
    Can be filters using the '--status' option, specifying
    pending, approved, or denied.
    The output printed includes:
        user id of a ZMQ credential, or the common name of a CSR
        remote address of the credential or csr
        status of the credential or cert (either APPROVED, DENIED, or PENDING)

    """

    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return

    output_view = []
    # try:
    #     pending_csrs = conn.server.vip.rpc.call(AUTH, "get_pending_csrs").get(
    #         timeout=4)
    #     for csr in pending_csrs:
    #         output_view.append(
    #             {
    #                 "entry": {
    #                     "user_id": csr["identity"],
    #                     "address": csr["remote_ip_address"],
    #                 },
    #                 "status": csr["status"],
    #             }
    #         )
    # except TimeoutError:
    #     print("Certs timed out")
    try:
        approved_certs = conn.server.vip.rpc.call(AUTH, "get_approved_authorizations").get(timeout=4)
        for value in approved_certs:
            output_view.append({"entry": value, "status": "APPROVED"})
    except TimeoutError:
        print("Approved credentials timed out")
    try:
        denied_certs = conn.server.vip.rpc.call(AUTH, "get_denied_authorizations").get(timeout=4)
        for value in denied_certs:
            output_view.append({"entry": value, "status": "DENIED"})
    except TimeoutError:
        print("Denied credentials timed out")
    try:
        pending_certs = conn.server.vip.rpc.call(AUTH, "get_pending_authorizations").get(timeout=4)
        for value in pending_certs:
            output_view.append({"entry": value, "status": "PENDING"})
    except TimeoutError:
        print("Pending credentials timed out")

    if not output_view:
        print("No remote certificates or credentials")
        return

    if opts.status == "approved":
        output_view = [output for output in output_view if output["status"] == "APPROVED"]

    elif opts.status == "denied":
        output_view = [output for output in output_view if output["status"] == "DENIED"]

    elif opts.status == "pending":
        output_view = [output for output in output_view if output["status"] == "PENDING"]

    elif opts.status is not None:
        _stdout.write("Invalid parameter. Please use 'approved', 'denied', 'pending', "
                      "or leave blank to list all.\n")
        return

    if len(output_view) == 0:
        print(f"No {opts.status} remote certificates or credentials")
        return
    for output in output_view:
        output["entry"] = {
            "user_id" if k == "identity" else "address" if "remote_ip_address" else k: v
            for k, v in output["entry"].items()
        }
        for value in output["entry"]:
            if not output["entry"][value]:
                output["entry"][value] = "-"
    userid_width = max(5, max(len(str(output["entry"]["user_id"])) for output in output_view))
    address_width = max(5, max(len(str(output["entry"]["address"])) for output in output_view))
    status_width = max(5, max(len(str(output["status"])) for output in output_view))
    fmt = "{:{}} {:{}} {:{}}\n"
    _stderr.write(fmt.format("USER_ID", userid_width, "ADDRESS", address_width, "STATUS", status_width))
    fmt = "{:{}} {:{}} {:{}}\n"
    for output in output_view:
        _stdout.write(
            fmt.format(
                output["entry"]["user_id"],
                userid_width,
                output["entry"]["address"],
                address_width,
                output["status"],
                status_width,
            ))


# TODO: Re-evaluate for usage
def approve_remote(opts):
    """Approves either a pending CSR or ZMQ credential.
    The platform must be running for this command to succeed.
    :param opts.user_id: The ZMQ credential user_id or pending CSR common name
    :type opts.user_id: str
    """
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    conn.server.vip.rpc.call(AUTH, "approve_authorization", opts.user_id).get(timeout=4)


# TODO: Re-evaluate for usage
def deny_remote(opts):
    """Denies either a pending CSR or ZMQ credential.
    The platform must be running for this command to succeed.
    :param opts.user_id: The ZMQ credential user_id or pending CSR common name
    :type opts.user_id: str
    """
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    conn.server.vip.rpc.call(AUTH, "deny_authorization", opts.user_id).get(timeout=4)


# TODO: Re-evaluate for usage
def delete_remote(opts):
    """Deletes either a pending CSR or ZMQ credential.
    The platform must be running for this command to succeed.
    :param opts.user_id: The ZMQ credential user_id or pending CSR common name
    :type opts.user_id: str
    """
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    conn.server.vip.rpc.call(AUTH, "delete_authorization", opts.user_id).get(timeout=4)


# def get_agent_publickey(opts):
#
#     def get_key(agent):
#         return opts.aip.get_agent_keystore(agent.uuid).public
#
#     _show_filtered_agents(opts, "PUBLICKEY", get_key)


# TODO: Re-evaluate for usage
def list_auth(opts, indices=None):
    _stdout.write("This method is under development\n")
    return

    # conn = opts.connection
    # if not conn:
    #     _stderr.write("VOLTTRON is not running. This command "
    #                   "requires VOLTTRON platform to be running\n")
    #     return
    #
    # entries = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["allow_list"]
    # print_out = []
    # if entries:
    #     for index, entry in enumerate(entries):
    #         if indices is None or index in indices:
    #             _stdout.write("\nINDEX: {}\n".format(index))
    #             _stdout.write("{}\n".format(jsonapi.dumps(entry, indent=2)))
    # else:
    #     _stdout.write("No entries in {}\n".format(os.path.join(get_home(), "auth.json")))


# TODO: Re-evaluate for usage
def _ask_for_auth_fields(
    domain=None,
    address=None,
    user_id=None,
    identity=None,
    capabilities=None,
    roles=None,
    groups=None,
    mechanism="CURVE",
    credentials=None,
    comments=None,
    enabled=True,
    **kwargs,
):
    """Prompts user for Auth Entry fields."""

    class Asker(object):

        def __init__(self):
            self._fields = collections.OrderedDict()

        def add(
                self,
                name,
                default=None,
                note=None,
                callback=lambda x: x,
                validate=lambda x, y: (True, ""),
        ):
            self._fields[name] = {
                "note": note,
                "default": default,
                "callback": callback,
                "validate": validate,
            }

        def ask(self):
            for name in self._fields:
                note = self._fields[name]["note"]
                default = self._fields[name]["default"]
                callback = self._fields[name]["callback"]
                validate = self._fields[name]["validate"]
                if isinstance(default, list):
                    default_str = "{}".format(",".join(default))
                elif default is None:
                    default_str = ""
                else:
                    default_str = default
                note = "({}) ".format(note) if note else ""
                question = "{} {}[{}]: ".format(name, note, default_str)
                valid = False
                while not valid:
                    response = input(question).strip()
                    if response == "":
                        response = default
                    if response == "clear":
                        if _ask_yes_no("Do you want to clear this field?"):
                            response = None
                    valid, msg = validate(response, self._fields)
                    if not valid:
                        _stderr.write("{}\n".format(msg))

                self._fields[name]["response"] = callback(response)
            return {k: self._fields[k]["response"] for k in self._fields}

    def to_true_or_false(response):
        if isinstance(response, str):
            return {"true": True, "false": False}[response.lower()]
        return response

    def is_true_or_false(x, fields):
        if x is not None:
            if isinstance(x, bool) or x.lower() in ["true", "false"]:
                return True, None
        return False, "Please enter True or False"

    def valid_creds(creds, fields):
        try:
            mechanism = fields["mechanism"]["response"]
            AuthEntry.valid_credentials(creds, mechanism=mechanism)
        except AuthException as e:
            return False, str(e)
        return True, None

    def valid_mech(mech, fields):
        try:
            AuthEntry.valid_mechanism(mech)
        except AuthException as e:
            return False, str(e)
        return True, None

    asker = Asker()
    asker.add("domain", domain)
    asker.add("address", address)
    asker.add("user_id", user_id)
    asker.add("identity", identity)
    asker.add(
        "capabilities",
        capabilities,
        "delimit multiple entries with comma",
        _parse_capabilities,
    )
    asker.add("roles", roles, "delimit multiple entries with comma", _comma_split)
    asker.add("groups", groups, "delimit multiple entries with comma", _comma_split)
    asker.add("mechanism", mechanism, validate=valid_mech)
    asker.add("credentials", credentials, validate=valid_creds)
    asker.add("comments", comments)
    asker.add("enabled", enabled, callback=to_true_or_false, validate=is_true_or_false)

    return asker.ask()


def _comma_split(line):
    if not isinstance(line, str):
        return line
    line = line.strip()
    if not line:
        return []
    return [word.strip() for word in line.split(",")]


def _parse_capabilities(line):
    if not isinstance(line, str):
        return line
    line = line.strip()
    try:
        result = jsonapi.loads(line.replace("'", '"'))
    except Exception as e:
        result = _comma_split(line)
    return result


def add_auth(opts):
    """Add authorization entry.

    If all options are None, then use interactive 'wizard.'
    """
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return

    fields = {
        "domain": opts.domain,
        "address": opts.address,
        "identity": opts.identity,
    #"comments": opts.comments,
    }

    if any(fields.values()):
        # Remove unspecified options so the default parameters are used
        fields = {k: v for k, v in fields.items() if v}
        #fields["enabled"] = not opts.disabled
        entry = fields
    else:
        # No options were specified, use interactive wizard
        responses = _ask_for_auth_fields()
        responses["rpc_method_authorizations"] = None
        entry = responses

    # if opts.add_known_host:
    #     if entry["address"] is None:
    #         raise ValueError("host (--address) is required when "
    #                          "--add-known-host is "
    #                          "specified")
    #     if entry["credentials"] is None:
    #         raise ValueError("serverkey (--credentials) is required when "
    #                          "--add-known-host is specified")
    #     opts.host = entry["address"]
    #     opts.serverkey = entry["credentials"]
    #     add_server_key(opts)

    try:
        # TODO: Add more than just identity maybe.
        value = conn.server.vip.rpc.call(AUTH, "create_credentials", identity=opts.identity).get(timeout=4)
        if value:
            _stdout.write("added credentials for {}\n".format(opts.identity))
        else:
            _stdout.write(f"Unable to add credentials for {opts.identity}\n")
    except AuthException as err:
        _stderr.write("ERROR: %s\n" % str(err))


def remove_auth(opts):
    from volttron.client.commands.control import _ask_yes_no
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    #
    # entry_count = len(conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["allow_list"])
    #
    # for i in opts.indices:
    #     if i < 0 or i >= entry_count:
    #         _stderr.write("ERROR: invalid index {}\n".format(i))
    #         return
    _stdout.write(f"This action will remove the identity {opts.identity}\n")

    if not _ask_yes_no("Do you wish to delete?"):
        return
    try:
        conn.server.vip.rpc.call(AUTH, "remove_credentials", identity=opts.identity)
        msg = f"{opts.identity} removed!"
        _stdout.write(msg + "\n")
    except AuthException as err:
        _stderr.write("ERROR: %s\n" % str(err))


def update_auth(opts):
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return

    entries = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["allow_list"]
    try:
        if opts.index < 0:
            raise IndexError
        entry = entries[opts.index]
        _stdout.write('(For any field type "clear" to clear the value.)\n')
        response = _ask_for_auth_fields(**entry)
        response["rpc_method_authorizations"] = None
        updated_entry = response
        conn.server.vip.rpc.call(AUTH, "auth_file.update_by_index", updated_entry, opts.index)
        _stdout.write("updated entry at index {}\n".format(opts.index))
    except IndexError:
        _stderr.write("ERROR: invalid index %s\n" % opts.index)
    except AuthException as err:
        _stderr.write("ERROR: %s\n" % str(err))


def add_role(opts):
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return

    roles = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["roles"]
    if opts.role in roles:
        _stderr.write('role "{}" already exists\n'.format(opts.role))
        return
    roles[opts.role] = list(set(opts.capabilities))
    conn.server.vip.rpc.call(AUTH, "auth_file.set_roles", roles)
    _stdout.write('added role "{}"\n'.format(opts.role))


def list_roles(opts):
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    roles = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["roles"]
    _print_two_columns(roles, "ROLE", "CAPABILITIES")


def update_role(opts):
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    roles = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["roles"]
    if opts.role not in roles:
        _stderr.write('role "{}" does not exist\n'.format(opts.role))
        return
    caps = roles[opts.role]
    if opts.remove:
        roles[opts.role] = list(set(caps) - set(opts.capabilities))
    else:
        roles[opts.role] = list(set(caps) | set(opts.capabilities))
    conn.server.vip.rpc.call(AUTH, "auth_file.set_roles", roles)
    _stdout.write('updated role "{}"\n'.format(opts.role))


def remove_role(opts):
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    roles = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["roles"]
    if opts.role not in roles:
        _stderr.write('role "{}" does not exist\n'.format(opts.role))
        return
    del roles[opts.role]
    conn.server.vip.rpc.call(AUTH, "auth_file.set_roles", roles)
    _stdout.write('removed role "{}"\n'.format(opts.role))


def add_group(opts):
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    groups = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["groups"]
    if opts.group in groups:
        _stderr.write('group "{}" already exists\n'.format(opts.group))
        return
    groups[opts.group] = list(set(opts.roles))
    conn.server.vip.rpc.call(AUTH, "auth_file.set_groups", groups)
    _stdout.write('added group "{}"\n'.format(opts.group))


def list_groups(opts):
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    groups = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["groups"]
    _print_two_columns(groups, "GROUPS", "ROLES")


def update_group(opts):
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    groups = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["groups"]
    if opts.group not in groups:
        _stderr.write('group "{}" does not exist\n'.format(opts.group))
        return
    roles = groups[opts.group]
    if opts.remove:
        groups[opts.group] = list(set(roles) - set(opts.roles))
    else:
        groups[opts.group] = list(set(roles) | set(opts.roles))
    conn.server.vip.rpc.call(AUTH, "auth_file.set_groups", groups)
    _stdout.write('updated group "{}"\n'.format(opts.group))


def remove_group(opts):
    conn = opts.connection
    if not conn:
        _stderr.write("VOLTTRON is not running. This command "
                      "requires VOLTTRON platform to be running\n")
        return
    groups = conn.server.vip.rpc.call(AUTH, "auth_file.read").get()["groups"]
    if opts.group not in groups:
        _stderr.write('group "{}" does not exist\n'.format(opts.group))
        return
    del groups[opts.group]
    conn.server.vip.rpc.call(AUTH, "auth_file.set_groups", groups)
    _stdout.write('removed group "{}"\n'.format(opts.group))


def add_agent_rpc_authorizations(opts):
    """
    Adds authorizations to method in auth entry in auth file.

    :param opts: Contains command line pattern and connection
    :return: None
    """
    conn = opts.connection
    agent_id = ".".join(opts.pattern[0].split(".")[:-1])
    agent_method = opts.pattern[0].split(".")[-1]
    if len(opts.pattern) < 2:
        _log.error("Missing authorizations for method. "
                   "Should be in the format agent_id.method "
                   "authorized_capability1 authorized_capability2 ...")
        return
    added_auths = [x for x in opts.pattern[1:]]
    try:
        conn.server.vip.rpc.call(AUTH, "add_rpc_authorizations", agent_id, agent_method, added_auths).get(timeout=4)
    except TimeoutError:
        _log.error(f"Adding RPC authorizations {added_auths} for {agent_id}'s "
                   f"method {agent_method} timed out")
    except Exception as e:
        _log.error(f"{e}) \nCommand format should be agent_id.method "
                   f"authorized_capability1 authorized_capability2 ...")
    return


def remove_agent_rpc_authorizations(opts):
    """
    Removes authorizations to method in auth entry in auth file.

    :param opts: Contains command line pattern and connection
    :return: None
    """
    conn = opts.connection
    agent_id = ".".join(opts.pattern[0].split(".")[:-1])
    agent_method = opts.pattern[0].split(".")[-1]
    if len(opts.pattern) < 2:
        _log.error("Missing authorizations for method. "
                   "Should be in the format agent_id.method "
                   "authorized_capability1 authorized_capability2 ...")
        return
    removed_auths = [x for x in opts.pattern[1:]]
    try:
        conn.server.vip.rpc.call(
            AUTH,
            "delete_rpc_authorizations",
            agent_id,
            agent_method,
            removed_auths,
        ).get(timeout=4)
    except TimeoutError:
        _log.error(f"Adding RPC authorizations {removed_auths} for {agent_id}'s "
                   f"method {agent_method} timed out")
    except Exception as e:
        _log.error(f"{e}) \nCommand format should be agent_id.method "
                   f"authorized_capability1 authorized_capability2 ...")
    return


def add_auth_parser(add_parser_fn, filterable):
    """
    This method should only be called if there is an auth service available otherwise
    this function will fail.

    :param add_parser_fn:
    :param filterable:
    :return:
    """
    auth_cmds = add_parser_fn("auth", help="manage agent credentials")

    auth_subparsers = auth_cmds.add_subparsers(title="subcommands", metavar="", dest="store_commands")

    auth_add = add_parser_fn("add", help="add new credentials", subparser=auth_subparsers)
    auth_add.add_argument("identity", help="Agent identity to add credentials for.")
    auth_add.add_argument("--domain", default=None)
    # auth_add.add_argument("--agent-address", default=None)
    # auth_add.add_argument("--mechanism", default=None)
    # auth_add.add_argument("--credentials", default=None)
    # auth_add.add_argument("--user_id", default=None)
    # auth_add.add_argument("--identity", default=None)
    #auth_add.add_argument("--groups", default=None, help="delimit multiple entries with comma")
    #auth_add.add_argument("--roles", default=None, help="delimit multiple entries with comma")
    # auth_add.add_argument("--capabilities",
    #                       default=None,
    #                       help="delimit multiple entries with comma")
    # auth_add.add_argument("--comments", default=None)
    # auth_add.add_argument("--disabled", action="store_true")
    # auth_add.add_argument("--add-known-host", action="store_true", help="adds entry in known host")
    auth_add.set_defaults(func=add_auth)

    # auth_add_group = add_parser_fn(
    #     "add-group",
    #     subparser=auth_subparsers,
    #     help="associate a group name with a set of roles",
    # )
    # auth_add_group.add_argument("group", metavar="GROUP", help="name of group")
    # auth_add_group.add_argument("roles",
    #                             metavar="ROLE",
    #                             nargs="*",
    #                             help="roles to associate with the group")
    # auth_add_group.set_defaults(func=add_group)
    #
    # auth_add_known_host = add_parser_fn(
    #     "add-known-host",
    #     subparser=auth_subparsers,
    #     help="add server public key to known-hosts file",
    # )
    # auth_add_known_host.add_argument("--host",
    #                                  required=True,
    #                                  help="hostname or IP address with optional port")
    # auth_add_known_host.add_argument("--serverkey", required=True)
    # auth_add_known_host.set_defaults(func=add_server_key)
    #
    # auth_add_role = add_parser_fn(
    #     "add-role",
    #     subparser=auth_subparsers,
    #     help="associate a role name with a set of capabilities",
    # )
    # auth_add_role.add_argument("role", metavar="ROLE", help="name of role")
    # auth_add_role.add_argument(
    #     "capabilities",
    #     metavar="CAPABILITY",
    #     nargs="*",
    #     help="capabilities to associate with the role",
    # )
    # auth_add_role.set_defaults(func=add_role)
    #
    # auth_keypair = add_parser_fn(
    #     "keypair",
    #     subparser=auth_subparsers,
    #     help="generate CurveMQ keys for encrypting VIP connections",
    # )
    # auth_keypair.set_defaults(func=gen_keypair)

    auth_list = add_parser_fn("list", help="list authentication records", subparser=auth_subparsers)
    auth_list.set_defaults(func=list_auth)

    # auth_list_groups = add_parser_fn(
    #     "list-groups",
    #     subparser=auth_subparsers,
    #     help="show list of group names and their sets of roles",
    # )
    # auth_list_groups.set_defaults(func=list_groups)
    #
    # auth_list_known_host = add_parser_fn(
    #     "list-known-hosts",
    #     subparser=auth_subparsers,
    #     help="list entries from known-hosts file",
    # )
    # auth_list_known_host.set_defaults(func=list_known_hosts)

    # auth_list_roles = add_parser_fn(
    #     "list-roles",
    #     subparser=auth_subparsers,
    #     help="show list of role names and their sets of capabilities",
    # )
    # auth_list_roles.set_defaults(func=list_roles)
    #
    # auth_publickey = add_parser_fn(
    #     "publickey",
    #     parents=[filterable],
    #     subparser=auth_subparsers,
    #     help="show public key for each agent",
    # )
    # auth_publickey.add_argument("pattern", nargs="*", help="UUID or name of agent")
    # auth_publickey.add_argument(
    #     "-n",
    #     dest="min_uuid_len",
    #     type=int,
    #     metavar="N",
    #     help="show at least N characters of UUID (0 to show all)",
    # )
    # auth_publickey.set_defaults(func=get_agent_publickey, min_uuid_len=1)

    auth_remove = add_parser_fn(
        "remove",
        subparser=auth_subparsers,
        help="removes one or more authentication records by indices",
    )
    auth_remove.add_argument("identity", help="Remove the identity from the credentials store.")
    # auth_remove.add_argument("indices",
    #                          nargs="+",
    #                          type=int,
    #                          help="index or indices of record(s) to remove")
    auth_remove.set_defaults(func=remove_auth)

    # auth_remove_group = add_parser_fn(
    #     "remove-group",
    #     subparser=auth_subparsers,
    #     help="disassociate a group name from a set of roles",
    # )
    # auth_remove_group.add_argument("group", help="name of group")
    # auth_remove_group.set_defaults(func=remove_group)
    #
    # auth_remove_known_host = add_parser_fn(
    #     "remove-known-host",
    #     subparser=auth_subparsers,
    #     help="remove entry from known-hosts file",
    # )
    # auth_remove_known_host.add_argument("host",
    #                                     metavar="HOST",
    #                                     help="hostname or IP address with optional port")
    # auth_remove_known_host.set_defaults(func=remove_known_host)

    # auth_remove_role = add_parser_fn(
    #     "remove-role",
    #     subparser=auth_subparsers,
    #     help="disassociate a role name from a set of capabilities",
    # )
    # auth_remove_role.add_argument("role", help="name of role")
    # auth_remove_role.set_defaults(func=remove_role)
    #
    # auth_serverkey = add_parser_fn(
    #     "serverkey",
    #     subparser=auth_subparsers,
    #     help="show the serverkey for the instance",
    # )
    # auth_serverkey.set_defaults(func=show_serverkey)

    # auth_update = add_parser_fn(
    #     "update",
    #     subparser=auth_subparsers,
    #     help="updates one authentication record by index",
    # )
    # auth_update.add_argument("index", type=int, help="index of record to update")
    # auth_update.set_defaults(func=update_auth)
    #
    # auth_update_group = add_parser_fn(
    #     "update-group",
    #     subparser=auth_subparsers,
    #     help="update group to include (or remove) given roles",
    # )
    # auth_update_group.add_argument("group", metavar="GROUP", help="name of group")
    # auth_update_group.add_argument(
    #     "roles",
    #     nargs="*",
    #     metavar="ROLE",
    #     help="roles to append to (or remove from) the group",
    # )
    # auth_update_group.add_argument("--remove",
    #                                action="store_true",
    #                                help="remove (rather than append) given roles")
    # auth_update_group.set_defaults(func=update_group)
    #
    # auth_update_role = add_parser_fn(
    #     "update-role",
    #     subparser=auth_subparsers,
    #     help="update role to include (or remove) given capabilities",
    # )
    # auth_update_role.add_argument("role", metavar="ROLE", help="name of role")
    # auth_update_role.add_argument(
    #     "capabilities",
    #     nargs="*",
    #     metavar="CAPABILITY",
    #     help="capabilities to append to (or remove from) the role",
    # )
    # auth_update_role.add_argument(
    #     "--remove",
    #     action="store_true",
    #     help="remove (rather than append) given capabilities",
    # )
    # auth_update_role.set_defaults(func=update_role)

    # auth_remote = add_parser_fn(
    #     "remote",
    #     subparser=auth_subparsers,
    #     help="manage pending RMQ certs and ZMQ credentials",
    # )
    # auth_remote_subparsers = auth_remote.add_subparsers(title="remote subcommands",
    #                                                     metavar="",
    #                                                     dest="store_commands")
    #
    # auth_remote_list_cmd = add_parser_fn(
    #     "list",
    #     subparser=auth_remote_subparsers,
    #     help="lists approved, denied, and pending certs and credentials",
    # )
    # auth_remote_list_cmd.add_argument("--status", help="Specify approved, denied, or pending")
    # auth_remote_list_cmd.set_defaults(func=list_remotes)
    #
    # auth_remote_approve_cmd = add_parser_fn(
    #     "approve",
    #     subparser=auth_remote_subparsers,
    #     help="approves pending or denied remote connection",
    # )
    # auth_remote_approve_cmd.add_argument(
    #     "user_id", help="user_id or identity of pending credential or cert to approve")
    # auth_remote_approve_cmd.set_defaults(func=approve_remote)
    #
    # auth_remote_deny_cmd = add_parser_fn(
    #     "deny",
    #     subparser=auth_remote_subparsers,
    #     help="denies pending or denied remote connection",
    # )
    # auth_remote_deny_cmd.add_argument(
    #     "user_id", help="user_id or identity of pending credential or cert to deny")
    # auth_remote_deny_cmd.set_defaults(func=deny_remote)
    #
    # auth_remote_delete_cmd = add_parser_fn(
    #     "delete",
    #     subparser=auth_remote_subparsers,
    #     help="approves pending or denied remote connection",
    # )
    # auth_remote_delete_cmd.add_argument(
    #     "user_id", help="user_id or identity of pending credential or cert to delete")
    # auth_remote_delete_cmd.set_defaults(func=delete_remote)

    # auth_rpc = add_parser_fn("rpc",
    #                          subparser=auth_subparsers,
    #                          help="Manage rpc method authorizations")
    #
    # auth_rpc_subparsers = auth_rpc.add_subparsers(title="subcommands",
    #                                               metavar="",
    #                                               dest="store_commands")
    # auth_rpc_add = add_parser_fn("add",
    #                              subparser=auth_rpc_subparsers,
    #                              help="adds rpc method authorizations")
    #
    # auth_rpc_add.add_argument(
    #     "pattern",
    #     nargs="*",
    #     help="Identity of agent and method, followed "
    #     "by capabilities. "
    #     "Should be in the format: "
    #     "agent_id.method authorized_capability1 "
    #     "authorized_capability2 ...",
    # )
    # auth_rpc_add.set_defaults(func=add_agent_rpc_authorizations, min_uuid_len=1)
    #
    # auth_rpc_remove = add_parser_fn(
    #     "remove",
    #     subparser=auth_rpc_subparsers,
    #     help="removes rpc method authorizations",
    # )
    #
    # auth_rpc_remove.add_argument(
    #     "pattern",
    #     nargs="*",
    #     help="Identity of agent and method, "
    #     "followed by capabilities. "
    #     "Should be in the format: "
    #     "agent_id.method "
    #     "authorized_capability1 "
    #     "authorized_capability2 ...",
    # )
    # auth_rpc_remove.set_defaults(func=remove_agent_rpc_authorizations, min_uuid_len=1)





def publish_to_bus(opts):
    print(f"Publishing to {opts.topic} {opts.data}")
    opts.connection.server.vip.pubsub.publish("pubsub", topic=opts.topic, message=opts.data).get()

def add_publish_parser(add_parser_fn):

    publisher = add_parser_fn("publish", help="Allow a single shot publisher to the bus from the command line.")
    
    #publisher_subparser = publisher.add_subparsers(title="publish options")
    
    publisher.add_argument("topic", help="Topic to publish to")
    publisher.add_argument("data", help="Data to publish")
    # publisher_subparser.add_argument("--internal-only", action="store_true"
    #                                  help="Should it be published internally only?")

    publisher.set_defaults(func=publish_to_bus)


# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import weakref

from .base import SubsystemBase
from ..results import ResultsDictionary

__all__ = ["Query"]


class Query(SubsystemBase):

    def __init__(self, core):
        self.core = weakref.ref(core)
        self._results = ResultsDictionary()
        core.register("query", self._handle_result, self._handle_error)

    def query(self, prop: str, peer: str = ""):
        """query a specific peer for a property value

        This method is very useful for retrieving configuration data from the core platform.  When
        peer is not specified it is defaulted to the router.

        :param prop:
            The property to query for.
        :param peer:
            The query to query upon
        :return:
        """
        connection = self.core().connection
        result = next(self._results)
        connection.send_vip(peer, "query", args=[prop], msg_id=result.ident)
        return result

    __call__ = query

    def _handle_result(self, message):
        if message.args and not message.args[0]:
            try:
                result = self._results.pop(message.id)
            except KeyError:
                return
            try:
                value = message.args[1]
            except IndexError:
                value = None
            result.set(value)

    def _handle_error(self, sender, message, error, **kwargs):
        try:
            result = self._results.pop(message.id)
        except KeyError:
            return
        result.set_exception(error)



"""
decorators.py
=============

This module contains a factory registration function and several instances of it. It also
includes a function for logging traces.
"""
import inspect
import logging
from typing import TypeVar

import gevent
from gevent import Greenlet

from volttron.client.known_identities import AUTH, CONFIGURATION_STORE
from volttron.server.containers import service_repo
from volttron.server.server_options import ServerOptions
from volttron.types.auth.auth_credentials import (Credentials, CredentialsCreator, CredentialsStore)
from volttron.types.auth.auth_service import (AuthService, Authenticator, AuthorizationManager, Authorizer)
from volttron.types import (AbstractAgent, AbstractCore, AgentBuilder, AgentExecutor, AgentStarter, MessageBus)



_log = logging.getLogger(__name__)

T = TypeVar('T')


def factory_registration(registy_name: str, protocol: T = None, singleton: bool = True, allow_many: bool = False):
    """
    Create a factory registration function.

    The function will have a registry attribute that is a dictionary of registered
    classes and a name attribute that is the name of the factory.

    @param name: The name of the factory.
    @type name: str
    @return: The factory registration function.
    @rtype: function
    """
    if protocol:
        _log.info(
            f"Setting registry for {registy_name} using protocol: {protocol.__name__} as singleton: {singleton} allowing many: {allow_many}"
        )
    else:
        _log.info(
            f"Setting registry for {registy_name} without protocol as singleton: {singleton} allowing many: {allow_many}"
        )

    def register(cls, **kwargs):
        lookup_key = None

        if hasattr(cls, 'Meta'):
            # Meta can either have a name or an identity, but not both.
            # We use either one of the values as a lookup for the register.
            if hasattr(cls.Meta, 'name') and hasattr(cls.Meta, 'identity'):
                raise ValueError("Only name or identity can be specified in Meta.")
            elif hasattr(cls.Meta, 'name'):
                lookup_key = cls.Meta.name
            elif hasattr(cls.Meta, 'identity'):
                lookup_key = cls.Meta.identity
        else:
            lookup_key = cls.__name__

        if lookup_key is None:
            raise ValueError(f"{cls.__name__} does not have an internal Meta class with identity or name.")

        # args = typing.get_args(protocol)
        # if
        if protocol is not None and not protocol in cls.__bases__ and not isinstance(cls, protocol):
            raise ValueError(f"{cls.__name__} doesn't implement {protocol}")

        if singleton:
            if allow_many:
                service_repo.add_concrete_reference(protocol, cls, kwargs=kwargs)
            else:
                service_repo.add_interface_reference(protocol, cls, kwargs=kwargs)
        else:
            service_repo.add_factory(cls, cls, kwargs=kwargs)

        if lookup_key is None:
            _log.warning("registration lookup key is None")
        else:
            _log.debug(f"Registering {cls.__name__} as a {lookup_key}")

        if lookup_key in register.registry:
            raise ValueError(f"{lookup_key} already in register for {register.name}.")
        register.registry[lookup_key] = cls
        return cls

    register.registy_name = registy_name
    register.registry = {}
    return register


factory = factory_registration("factories")

# Allow many so lookup based upon concrete class rather than interface.
service = factory_registration("service", singleton=True, allow_many=True)

credentials_store = factory_registration("credentials_store", protocol=CredentialsStore)
credentials_creator = factory_registration("credentials_creator", protocol=CredentialsCreator)

# core = factory_registration("core", protocol=Core, singleton=False)
# connection = factory_registration("connection", protocol=Connection | ConnectionBuilder)
messagebus = factory_registration("messagebus", protocol=MessageBus)

agent_starter = factory_registration("agent_starter", protocol=AgentStarter)
agent_executor = factory_registration("agent_executor", protocol=AgentExecutor)
agent_builder = factory_registration("agent_builder", protocol=AgentBuilder)

authservice = factory_registration("authservice", protocol=AuthService)

authorizer = factory_registration("authorizer", protocol=Authorizer)
authenticator = factory_registration("authenticator", protocol=Authenticator)
authorization_manager = factory_registration("authorization_manager", protocol=AuthorizationManager)

auth_create_hook = factory_registration("auth_create_hook")
auth_add_hook = factory_registration("auth_add_hook")
auth_remove_hook = factory_registration("auth_remove_hook")
auth_list_hook = factory_registration("auth_list_hook")


def __get_create_instance_from_factory__(*, instances, registration, name: str = None, **kwargs):
    if not registration.registry:
        raise ValueError(f"No {registration.registy_name} is currently registered")

    if name is None and len(registration.registry) > 1:
        raise ValueError(f"Can't figure out which messagebus to return.")

    the_instance = None
    if name is None:
        # First name of the registry dictionary.
        name = list(registration.registry.keys())[0]
        signature = inspect.signature(registration.registry[name].__init__)
        for k, v in kwargs.items():
            if k not in signature.parameters:
                raise ValueError(f"Invalid parameter {k} for {name} signature has {signature.parameters}")

        instances[name] = registration.registry[name](**kwargs)
    elif name not in instances:
        instances[name] = registration.registry[name]()

    the_instance = instances.get(name)
    if the_instance is None:
        raise ValueError(f"Couldn't retrieve {name} from register")
    return the_instance


def __get_class_from_factory__(registration, name: str = None):
    if not registration.registry:
        raise ValueError(f"No {name} is currently registered")

    if name is None and len(registration.registry) > 1:
        raise ValueError(f"Can't figure out which messagebus to return.")

    if name is None:
        # First name of the registry dictionary.
        name = list(registration.registry.keys())[0]

    if name not in registration.registry:
        raise ValueError(f"Couldn't retrieve {name} from register")

    return registration.registry.get(name)


__messagebus__: dict[str, object] = {}


def get_messagebus_instance(name=None) -> MessageBus:
    return __get_create_instance_from_factory__(__messagebus__, messagebus, name)


__messagebus_core__: dict[str, object] = {}


def get_messagebus_core(name=None) -> object:
    return __get_create_instance_from_factory__(__messagebus_core__, core, name)


__messagebus_connection__: dict[str, object] = {}


def get_messagebus_core(name=None) -> object:
    """
    Return a messagebus connection instance.

    :param name: _description_, defaults to None
    :type name: _type_, optional
    :return: _description_
    :rtype: object
    """
    return __get_create_instance_from_factory__(__messagebus_connection__, connection, name)


def get_messagebus_class(name: str = None) -> type:
    """
    Return a registered messagebus class.

    :param name: The name of the registered messagebus class, defaults to None
    :type name: str, optional
    :return: The registered messagebus class
    :rtype: type
    :raises ValueError: If no messagebus class is registered or if the passed name doesn't
        exist in the registry
    """
    return __get_class_from_factory__(messagebus, name)


def get_core_instance(credentials: Credentials) -> AbstractCore:
    """
    Return a registered core class.

    :param name: The name of the registered core class, defaults to None
    :type name: str, optional
    :return: The registered core class
    :rtype: type
    :raises ValueError: If no core class is registered or if the passed name doesn't
        exist in the registry
    """
    registerd_cls = __get_class_from_factory__(core)
    service_repo.add_instance(Credentials, credentials)

    return service_repo.resolve(registerd_cls)
    # return __get_class_from_factory__(core, name)


__authorizer__: dict[str, Authorizer] = {}


def get_authorizer(name: str = None, authorization_manager: AuthorizationManager = None) -> Authorizer:
    authorizer_instance: Authorizer = None
    if name is not None:
        authorizeritem = __authorizer__.get(name, None)

    # Use the default authorization manager if none is provided.
    if authorization_manager is None:
        authorization_manager = get_authorization_manager()

    assert isinstance(authorization_manager, AuthorizationManager)

    if authorizer_instance is None:
        authorizer_instance = __get_create_instance_from_factory__(instances=__authorizer__,
                                                                   registration=authorizer,
                                                                   name=name,
                                                                   authorization_manager=authorization_manager)
        if authorizer_instance is not None:
            __authorizer__[name] = authorizer_instance
    return authorizer_instance


__authenticator__: dict[str, Authenticator] = {}


def get_authenticator(name: str = None, credentials_creator: CredentialsCreator = None) -> Authenticator:
    authenticatoritem: Authenticator = None
    if name is not None:
        authenticatoritem = __authenticator__.get(name, None)
    if credentials_creator is None:
        credentials_creator = get_credentials_creator()

    assert isinstance(credentials_creator, CredentialsCreator)

    if authenticatoritem is None:
        authenticatoritem = __get_create_instance_from_factory__(instances=__authenticator__,
                                                                 registration=authenticator,
                                                                 name=name,
                                                                 credentials_creator=credentials_creator)
        if authenticatoritem is not None:
            __authenticator__[name] = authenticatoritem
    return authenticatoritem


__credentials_store__: dict[str, CredentialsStore] = {}


def get_credentials_store(name: str = None) -> CredentialsStore:
    credentials_store_item: CredentialsStore = None
    if name is not None:
        credentials_store_item = __credentials_store__.get(name, None)
    if credentials_store_item is None:
        credentials_store_item = __get_create_instance_from_factory__(instances=__credentials_store__,
                                                                      registration=credentials_store,
                                                                      name=name)
        if credentials_store_item is not None:
            __credentials_store__[name] = credentials_store_item
    return credentials_store_item


__authorization_manager__: dict[str, AuthorizationManager] = {}


def get_authorization_manager(name: str = None) -> AuthorizationManager:
    auth_manager: AuthorizationManager = None
    if name is not None:
        auth_manager = __authorization_manager__.get(name, None)
    if auth_manager is None:
        auth_manager = __get_create_instance_from_factory__(instances=__authorization_manager__,
                                                            registration=authorization_manager,
                                                            name=name)
        if auth_manager is not None:
            __authorization_manager__[name] = auth_manager
    return auth_manager


def get_authservice_class(name=None) -> type:
    return __get_class_from_factory__(authservice, name)


__credentials_creator__: dict[str, CredentialsCreator] = {}


def get_credentials_creator(name=None) -> CredentialsCreator:
    creator_item: CredentialsCreator = None
    if name is not None:
        creator_item = __credentials_creator__.get(name, None)
    if creator_item is None:
        creator_item = __get_create_instance_from_factory__(instances=__credentials_creator__,
                                                            registration=credentials_creator,
                                                            name=name)
        if creator_item is not None:
            __credentials_creator__[name] = creator_item
    return creator_item


def get_services() -> dict[str, type]:
    return service.registry


def get_services_without_requires() -> list[type]:
    return list(filter(lambda x: not hasattr(x.Meta, "requires"), service.registry.values()))


def get_services_with_requires() -> list[type]:
    return list(filter(lambda x: hasattr(x.Meta, "requires"), service.registry.values()))


def get_service_class(identity: str) -> type:
    return service.registry[identity]


__service_instances__: dict[str, object] = {}


def get_service_instance(identity: str) -> object:
    return __get_create_instance_from_factory__(instances=__service_instances__, registration=service, name=identity)


def get_service_startup_order() -> list[str]:
    ordered: list[str] = ["platform.config"]

    for lookup, cls in service.registry.items():
        if lookup not in ('platform.config', ):
            ordered.append(lookup)

    return ordered


def start_service_agents() -> list[Greenlet]:
    """
    Start all services in the registry.

    :return: A list of greenlets that are running the services.
    :rtype: list[Greenlet]
    """
    _log.debug("Starting Service Agents")
    from volttron.types.auth.auth_service import AuthService

    greenlets = []
    options = service_repo.resolve(ServerOptions)

    for lookup, cls in service.registry.items():
        if isinstance(cls, type) and issubclass(cls, AbstractAgent) and cls.Meta.identity == AUTH:
            obj = service_repo.resolve(cls)
            event = gevent.event.Event()
            task = gevent.spawn(obj.core.run, event)
            event.wait()
            del event
            greenlets.append(task)
            break

    for lookup, cls in service.registry.items():
        if isinstance(cls, type) and issubclass(cls,
                                                AbstractAgent) and cls.Meta.identity not in (CONFIGURATION_STORE, AUTH):
            _log.debug(f"- {cls.__name__}")
            obj = service_repo.resolve(cls)
            event = gevent.event.Event()
            task = gevent.spawn(obj.core.run, event)
            event.wait()
            del event
            greenlets.append(task)
    _log.debug("End Starting Service Agents")
    return greenlets


#    is_required_by: dict[str, list[str]] = {}

# for k, r in service.registry.items():
#     is_required_by[k] = r

# for k, r in service.registry.items():
#     if hasattr(r, "Meta") and hasattr(r.Meta, "requires"):
#         if isinstance(r.Meta.requires, str):
#             r.Meta.requires = [r.Meta.requires]

#         for require in r.Meta.requires:
#             is_required_by[require].append(k)



from abc import ABC, abstractmethod
from typing import Dict, List, Any, Optional, Callable

class FederationBridge(ABC):
    """
    Abstract interface for platform federation bridges.
    
    A FederationBridge enables communication between separate VOLTTRON platforms.
    It handles the details of establishing connections, routing messages, and
    monitoring connection health between federated platforms.
    
    Each message bus implementation (ZMQ, RMQ, etc.) provides its own 
    implementation of this interface, allowing the Federation Service to work with
    any message bus without knowing the implementation details.
    """
    
    @abstractmethod
    def connect(self, platform_id: str, platform_address: str, credentials: Any) -> bool:
        """
        Connect to a remote platform
        
        :param platform_id: ID of the remote platform
        :param platform_address: Address of the remote platform
        :param credentials: Authentication credentials for the platform
        :return: True if connection was successful
        """
        pass
        
    @abstractmethod
    def disconnect(self, platform_id: str) -> bool:
        """
        Disconnect from a remote platform
        
        :param platform_id: ID of the remote platform
        :return: True if disconnection was successful
        """
        pass
        
    @abstractmethod
    def get_status(self, platform_id: Optional[str] = None) -> Dict[str, Any]:
        """
        Get status of federation bridges
        
        :param platform_id: Optional ID to get status for a specific platform
        :return: Dictionary with status information
        """
        pass
        
    @abstractmethod
    def ping(self, platform_id: str, timeout: int = 5) -> bool:
        """
        Ping a remote platform to check connection health
        
        :param platform_id: ID of the remote platform
        :param timeout: Timeout in seconds
        :return: True if ping was successful
        """
        pass


from __future__ import annotations

import importlib
import inspect
import logging
import pkgutil
import sys
import traceback
from pathlib import Path
from typing import List, Type
import ast
import os

__all__: List[str] = ["load_dir", "load_subclasses", "find_subpackages"]

_log = logging.getLogger(__name__)

__loaded__: set[str] = set()


def is_subclass(node, base_class_name):
    # Check if the node is a class that derives from the given base class
    return (
            isinstance(node, ast.ClassDef) and
            any(base.id == base_class_name for base in node.bases if isinstance(base, ast.Name))
    )


def find_subpackages(base_package_name):
    subpackages = set()

    # Iterate over every path in sys.path
    for path in sys.path:
        if not path or not isinstance(path, str):
            continue

        potential_package_path = os.path.join(path, *base_package_name.split('.'))

        if os.path.isdir(potential_package_path):
            for root, dirs, files in os.walk(potential_package_path):
                # Determine if it's a package by checking for __init__.py presence
                if '__init__.py' in files:
                    # Calculate the relative module path
                    rel_path = os.path.relpath(root, path)
                    package_name = rel_path.replace(os.sep, '.')
                    subpackages.add(package_name)

    return subpackages


def load_subclasses(base_class_with_package, package_prefix):
    """
    Given a base class (ex. volttron.types.MessageBus) return the concrete subclass available in
    path(example ZMQMessageBus).
    This assumes that the subclass in a package package_prefix.* For example ZMQMessageBus should be defined in a
    module within some package that starts with volttron.messagebus
    The code uses static loading to inspect all existing modules under package_prefix.* and imports only the
    module containing the subclasses of given base class. This is done to avoid importing all modules in the packages
    there by avoiding modules with blocking top level code such as while True: loop

    :param base_class_with_package: Baseclass name with package. Example: volttron.types.MessageBus
    :param package_prefix: prefix of package where to expect concrete implementation. Example: volttron.messagebus
    :return: set of subclasses
    """

    #import the base class
    module = base_class_with_package[:base_class_with_package.rindex(".")]
    m = importlib.import_module(module)
    base_class = getattr(m, base_class_with_package[base_class_with_package.rindex(".") + 1:])

    package = importlib.import_module(package_prefix)
    # loop through package_name prefix subpackages from all loaded libraries
    # for example, if package_prefix is volttron.auth package_path could return path to volttron.auth in both
    # volttron-lib-auth and volttron-lib-zmq
    base_class_name = base_class.__name__
    all_subclasses = set()
    for filefinder, subpackage_name, is_pkg in pkgutil.walk_packages(package.__path__, package.__name__ + "."):
        if is_pkg:
            module_name = subpackage_name
            file_path = os.path.join(filefinder.path, "__init__.py")
            if os.path.isfile(file_path):
                all_subclasses.union(inspect_module_for_subclasses(module_name, file_path, base_class, base_class_name))
        else:
            module_name = subpackage_name
            file_path = os.path.join(filefinder.path, subpackage_name.split('.')[-1] + ".py")
            all_subclasses.union(inspect_module_for_subclasses(module_name, file_path, base_class, base_class_name))

    return all_subclasses


def inspect_module_for_subclasses(module_name, file_path, base_class, base_class_name):
    subclasses = set()
    with open(file_path, 'r') as source_file:
        try:
            tree = ast.parse(source_file.read(), filename=file_path)

            for node in ast.walk(tree):
                if is_subclass(node, base_class_name):
                    module = importlib.import_module(module_name)
                    for name, obj in inspect.getmembers(module, inspect.isclass):
                        if issubclass(obj, base_class) and obj is not base_class:
                            subclasses.add(obj)
        except SyntaxError as e:
            _log.error(f"Syntax error in file {file_path}: {e}")
            raise e
        return subclasses


def load_dir(package: str, pth: Path):
    """
    Recursively loads all modules within a directory.

    :param package: The package name.
    :type package: str
    :param pth: The path to the directory.
    :type pth: :class:`pathlib.Path`

    This function recursively loads all modules within the specified directory. It iterates
    over the files and subdirectories in the directory, imports each module, and loads any
    subdirectories as sub-packages.

    If a file is a Python source file (`.py`), it is imported as a module. If a file is a
    directory, it is loaded as a sub-package.

    Example usage::

        load_dir('volttron', Path('/path/to/directory'))

    .. note::
        This function does not load compiled Python files (`.pyc`) or files in
        the `__pycache__` directory.

    .. warning::
        Be cautious when using this function, as it imports and loads all modules within a
        directory, which may have unintended side effects.
    """

    # Find the caller of this function
    caller_module = inspect.currentframe().f_back.f_globals["__name__"]

    get_mod_name = lambda pth: pth.name if pth.is_dir() else pth.stem

    # We only want to load from py files rather than pyc files
    # so filter
    for p in filter(lambda x: x.name != '__pycache__', pth.iterdir()):

        new_package = f"{package}.{get_mod_name(p)}"

        if new_package.endswith("__init__") or new_package.endswith('.mypy_cache'):
            _log.debug(f"Skipping {new_package}")
            continue

        if f"{caller_module}.__init__" == new_package:
            _log.debug(f"Skipping {new_package}")
            continue
        # print(globals())
        if p.absolute().as_posix() not in __loaded__:
            __loaded__.add(p.absolute().as_posix())
            _log.debug(f"Loading {new_package}")
            try:
                importlib.import_module(new_package)
            except ImportError as ex:
                traceback.print_exc()
                _log.error(f"Failed to import {new_package} due to {ex}")
                raise
                continue
            # after = set(globals().keys())
            # print("After import")
            # print(before - after)
            # print("AFTER!!!!")
            # for k in list(globals().keys()):
            #     print(k)

            # assert new_package in globals()
            if p.is_dir():
                load_dir(new_package, p)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import errno
import volttron.utils.jsonrpc as jsonrpc

__all__ = ["VIPError", "Unreachable", "Again", "UnknownSubsystem"]


class VIPError(Exception):

    def __init__(self, errnum, msg, peer, subsystem, *args):
        super(VIPError, self).__init__(errnum, msg, peer, subsystem, *args)
        self.errno = int(errnum)
        self.msg = msg
        self.peer = peer
        self.subsystem = subsystem

    def __str__(self):
        return "VIP error (%d): %s" % (self.errno, self.msg)

    def __repr__(self):
        return "%s%r" % (type(self).__name__, self.args)

    @classmethod
    def from_errno(cls, errnum, msg, *args):
        try:
            if isinstance(errnum, str):
                errnum = getattr(errno, errnum.split('.')[-1])

        except ValueError:
            return cls(999, msg, *args)
        return {
            errno.EHOSTUNREACH: Unreachable,
            errno.EAGAIN: Again,
            errno.EPROTONOSUPPORT: UnknownSubsystem,
            jsonrpc.UNAUTHORIZED: Unauthorized
        }.get(errnum, cls)(errnum, msg, *args)


class Unreachable(VIPError):

    def __str__(self):
        return "%s: %s" % (super(Unreachable, self).__str__(), self.peer)


class Again(VIPError):
    pass


class UnknownSubsystem(VIPError):

    def __str__(self):
        return "%s: %s" % (super(UnknownSubsystem, self).__str__(), self.subsystem)


class Unauthorized(VIPError):
    pass



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
"""The volttron.utils package contains generic utilities for handling json, storing configurations math
libraries...and more. """

from contextlib import contextmanager
from copy import deepcopy
import inspect
import logging
import os
from pathlib import Path
from typing import List, TYPE_CHECKING

import psutil
import yaml

from volttron.utils.commands import (is_volttron_running, execute_command, isapipe, wait_for_volttron_startup,
                                     wait_for_volttron_shutdown, vip_main)
from volttron.utils.context import ClientContext
from volttron.utils.commands import wait_for_volttron_startup, wait_for_volttron_shutdown
from volttron.utils.dynamic_helper import get_module, get_class, get_subclasses
from volttron.utils.file_access import create_file_if_missing

from volttron.utils.identities import normalize_identity, is_valid_identity
from volttron.utils.jsonapi import strip_comments, parse_json_config
from volttron.utils.messagebus import store_message_bus_config
from volttron.utils.network import get_hostname, is_ip_private
from volttron.utils.time import (format_timestamp, process_timestamp, parse_timestamp_string,
                                 get_utc_seconds_from_epoch, get_aware_utc_now, fix_sqlite3_datetime)
from volttron.utils.version import get_version
from volttron.types import Identity


# def get_logger() -> logging.Logger:
#     frame = inspect.stack()[1]
#     module = inspect.getmodule(frame[0])
#     return logging.getLogger(module.__name__)


_log = logging.getLogger(__name__)


@contextmanager
def set_agent_identity(identity: Identity):
    """
    A context manager allowing setting and unsetting of the AGENT_VIP_IDENTITY.

    :param identity: the identity to set
    :type identity: Identity
    """
    env_key = "AGENT_VIP_IDENTITY"
    before = os.environ.get(env_key)
    os.environ[env_key] = identity
    yield
    if before is not None:
        os.environ[env_key] = identity


def monkey_patch():
    from gevent import monkey

    # At this point these are the only things that need to be patched
    # and the server and client are working harmoniously with this.
    patches = [
        ('ssl', monkey.patch_ssl),
        ('socket', monkey.patch_socket),
        ('os', monkey.patch_os),
    ]

    # patch modules if necessary.  Only if the module hasn't been patched before.
    # this could happen if the server code uses the client (which it does).
    for module, fn in patches:
        if not monkey.is_module_patched(module):
            fn()


def is_regex(value: str) -> bool:
    """
    Determine if the passed value is a regular expression or not.  A regular expression
    is something that starts and ends with /, is a string and can be compiled.

    :param value: A potential regular expression
    :return: True if a volttron regular expression.
    """
    return value is not None and isinstance(value, str) and len(value) > 1 and value[0] == value[-1] == "/"


def load_config(default_configuration: str | Path | dict | None) -> dict:
    """
    Load the default configuration from a JSON or YAML encoded file.

    If default_configuration is None then return {}.

    If default_configuration is a dictionary, then return a deep copy of the dictionary
    without any changes to it.

    If default_configuration is a string or Path object, they must resolve to a file that
    is readable by the current process.  The file referenced will be parsed and loaded
    using yaml.safe_load.  Doing so will load both json or yaml based files.

    :param default_configuration: An agent configuration that is passed to __init__
    :type default_configuration: str | Path | dict
    :raises ValueError: If default_configuration is not resolvable.
    :return: A dictionary of the parsed default_configuration.
    :rtype: dict
    """

    if default_configuration is None or default_configuration == "":
        return {}

    if isinstance(default_configuration, dict):
        return deepcopy(default_configuration)

    if isinstance(default_configuration, str):
        default_configuration = Path(default_configuration).expanduser().absolute()
    elif isinstance(default_configuration, Path):
        default_configuration = default_configuration.expanduser().absolute()
    else:
        ValueError(
            f"Invalid type passed as default_configuration {type(default_configuration)} MUST be str | Path | dict | None"
        )

    # First attempt parsing the file with a yaml parser (allows comments natively)
    # Then if that fails we fallback to our modified json parser.
    try:
        return yaml.safe_load(default_configuration.read_text())
    except yaml.YAMLError as e:
        try:
            return parse_json_config(default_configuration.read_text())
        except Exception as e:
            _log.error("Problem parsing agent configuration")
            raise


def update_kwargs_with_config(kwargs, config):
    """
    Loads the user defined configurations into kwargs and converts any dash/hyphen in config variables into underscores
    :param kwargs: kwargs to be updated
    :param config: dictionary of user/agent configuration
    """

    for k, v in config.items():
        kwargs[k.replace("-", "_")] = v

def is_volttron_running(volttron_home):
    """
    Checks if volttron is running for the given volttron home. Checks if a VOLTTRON_PID file exist and if it does
    check if the PID in the file corresponds to a running process. If so, returns True else returns False
    :param vhome: volttron home
    :return: True if VOLTTRON_PID file exists and points to a valid process id
    """

    pid_file = os.path.join(volttron_home, 'VOLTTRON_PID')
    if os.path.exists(pid_file):
        running = False
        import builtins
        with builtins.open(pid_file, 'r') as pf:
            pid = int(pf.read().strip())
            running = psutil.pid_exists(pid)
        return running
    else:
        return False

__all__: List[str] = [
    "update_kwargs_with_config", "load_config", "parse_json_config", "get_hostname", "strip_comments", "is_regex",
    "is_valid_identity", "isapipe", "is_volttron_running", "create_file_if_missing", "wait_for_volttron_shutdown",
    "process_timestamp", "parse_timestamp_string", "execute_command", "get_version", "get_aware_utc_now",
    "get_utc_seconds_from_epoch", "wait_for_volttron_startup", "normalize_identity", "ClientContext",
    "format_timestamp", "store_message_bus_config", "is_ip_private", "fix_sqlite3_datetime", "vip_main", "get_module",
    "get_class", "get_subclasses", "monkey_patch", "is_volttron_running"
]



import argparse
import configparser
import logging
import os
import socket
from collections import OrderedDict
from configparser import ConfigParser
from dataclasses import dataclass, field, fields
from pathlib import Path

_log = logging.getLogger(__name__)

class MultiOrderedDict(OrderedDict):

    def __setitem__(self, key, value):
        if isinstance(value, list) and key in self:
            self[key].extend(value)
        else:
            super(MultiOrderedDict, self).__setitem__(key, value)


@dataclass
class ServerOptions:
    """
    A data class representing the configuration options for a Volttron platform server.

    :ivar volttron_home: The path to the root directory of the Volttron instance.
                         Default is '~/.volttron', which is expanded to the current user's home directory.
    :vartype volttron_home: Union[pathlib.Path, str]

    :ivar instance_name: The name of the Volttron instance. Default is the hostname of the machine.
    :vartype instance_name: str

    :ivar address: A list of addresses on which the platform should listen for incoming connections.
                     Default is None.
    :vartype address: List[str]

    :ivar agent_isolation_mode: Flag indicating whether the agent isolation mode is enabled.
                                Default is False.
    :vartype agent_isolation_mode: bool

    :ivar message_bus: The fully-qualified name of the message bus class to use. Default is
                       'volttron.messagebus.zmq.ZmqMessageBus'.
    :vartype message_bus: str

    :ivar agent_core: The fully-qualified name of the agent core class to use. Default is
                      'volttron.messagebus.zmq.ZmqCore'.
    :vartype agent_core: str

    :ivar auth_service: The fully-qualified name of the authentication service class to use. Default is
                        'volttron.services.auth'.
    :vartype auth_service: str

    :ivar service_config: The Path to the service config file for loading services into the context.
    :vartype service_service: Path
    """
    volttron_home: Path | None = None
    instance_name: str | None = None
    local_address: str | None = None
    address: list[str] = field(default_factory=list)
    agent_isolation_mode: bool = False
    # Module that holds the zmq based classes, though we shorten it assuming
    # it's in volttron.messagebus
    messagebus: str = "zmq"
    auth_enabled: bool = True
    config_file: Path | None = None
    initialized: bool = False
    service_address: str | None = None
    server_messagebus_id: str = "vip.server"
    agent_monitor_frequency: int = 30
    poetry_project_path: Path | None = None
    enable_federation: bool = False
    federation_url: str | None  = None
    

    def __post_init__(self):
        """
        Initializes the instance after it has been created.

        If `volttron_home` is a string, it is converted to a `pathlib.Path` object.

        If `instance_name` is None, it is set to the hostname of the machine.
        """

        if self.volttron_home is None:
            self.volttron_home = Path(os.environ.get("VOLTTRON_HOME", "~/.volttron")).expanduser()

        if self.poetry_project_path is None:
            self.poetry_project_path = self.volttron_home

        if isinstance(self.volttron_home, str):
            self.volttron_home = Path(self.volttron_home).absolute()
            
        # Should be the only location where we create VOLTTRON_HOME
        if not self.volttron_home.is_dir():
            self.volttron_home.mkdir(mode=0o755, exist_ok=True, parents=True)

        if self.config_file is None:
            self.config_file = self.volttron_home / "config"

        # TODO: This should be removed once we have a better way of handling
        if self.service_address is None:
            self.service_address = "inproc://vip"

        # Allow the config path to be whereever the user wants it to be.
        if not self.config_file.exists():
            self.config_file.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
            if self.instance_name is None:
                self.instance_name = socket.gethostname()

            if isinstance(self.address, str):
                self.address = [self.address]

        else:
            if not self.initialized:
                options = ServerOptions.from_file(self.config_file)

                for fld in ServerOptions.__dataclass_fields__:
                    setattr(self, fld, getattr(options, fld))

    from volttron.types import MessageBusConfig
    def get_messagebus_config(self) -> MessageBusConfig:
        """
        Get messagebus configuration using the global registry system
        
        Returns:
            MessageBusConfig: Configured messagebus instance
        """
        from volttron.types.messagebus import get_messagebus_config_class
        
        # Get messagebus type from config file or default to zmq
        messagebus_type = self.messagebus
        if not messagebus_type:
            # Check config file for messagebus setting
            config_file = self.volttron_home / "config"
            if config_file.exists():
                from configparser import ConfigParser
                config = ConfigParser()
                config.read(config_file)
                if config.has_option("volttron", "messagebus"):
                    messagebus_type = config.get("volttron", "messagebus")
            
            # Default to zmq if nothing specified
            messagebus_type = messagebus_type or "zmq"
        
        # Load config class from global registry
        config_class = get_messagebus_config_class(messagebus_type)
        if not config_class:
            raise ValueError(
                f"Unknown messagebus type '{messagebus_type}'. "
                f"Make sure the appropriate messagebus library is installed and registered."
            )
        
        # Create configuration with server options
        options = {
            "instance_name": self.instance_name,
            "volttron_home": str(self.volttron_home),
            "address": self.address,
            "local_address": self.local_address,
            "enable_federation": self.enable_federation,
            "federation_url": self.federation_url,
            "auth_enabled": self.auth_enabled,
            "service_address": self.service_address,
            "agent_monitor_frequency": self.agent_monitor_frequency
        }

    # volttron_home: Path | None = None
    # instance_name: str | None = None
    # local_address: str | None = None
    # address: list[str] = field(default_factory=list)
    # agent_isolation_mode: bool = False
    # # Module that holds the zmq based classes, though we shorten it assuming
    # # it's in volttron.messagebus
    # messagebus: str = "zmq"
    # auth_enabled: bool = True
    # config_file: Path | None = None
    # initialized: bool = False
    # service_address: str | None = None
    # server_messagebus_id: str = "vip.server"
    # agent_monitor_frequency: int = 30
    # poetry_project_path: Path | None = None
    # enable_federation: bool = False
    # federation_url: str | None  = None
        
        return config_class.create_from_options(options)
    
    def update(self, opts: argparse.Namespace | dict):
        """Update the opts from the passed command line or a dictionary.

        :param opts: Parameters passed from the command line or a dictionary form volttron testing framework.
        :type opts: argparse.Namespace | dict
        """
        address = set(opts.address)
        opts.address = list(address)

        if isinstance(opts, dict):
            dev_mode = opts.pop("dev_mode")
            self.__dict__.update(opts)
        else:
            d = opts.__dict__
            dev_mode = d.pop("dev_mode")
            self.__dict__.update(d)

        if dev_mode:
            self.poetry_project_path = Path(os.path.abspath(os.curdir))

    def store(self, file: Path = None):
        """
        Stores the current configuration options to a file.

        :param file: The path to the file where the configuration options should be stored.
        :type file: Union[pathlib.Path, str]
        """
        parser = ConfigParser(dict_type=MultiOrderedDict, strict=False)

        parser.add_section("volttron")

        kwargs = {}

        services_field = None
        # Store the config options first.
        for field in fields(ServerOptions):
            try:
                # Don't save volttron_home within the config file.
                if field.name not in ('volttron_home', 'services', 'config_file', 'initialized', 'service_address',
                                      "poetry_project_path", "local_address"):
                    # More than one address can be present, so we must be careful
                    # with it.
                    if field.name == 'address':
                        found = set()
                        for v in getattr(self, field.name):
                            if v not in found:
                                parser.set("volttron", "address", value=v)
                                found.add(v)
                        # for v in getattr(self, field.name):
                        #     parser.set("volttron", "address", v)
                    else:
                        parser.set("volttron", field.name.replace('_', '-'), str(getattr(self, field.name)))
            except configparser.NoOptionError:
                pass

        # TODO Add services back in.
        # parser.add_section('services')
        # for sd in self.services:
        #     parser.set("services", sd.identity, sd.klass_path)

        #     if sd.args:
        #         parser.add_section(sd.klass_path)
        #         for arg, value in sd.args:
        #             parser.set(sd.klass_path, arg, value)

        if file is None:
            file = self.config_file
        parser.write(file.open("w"))

    @staticmethod
    def from_file(file: Path | str = None):
        """
        Creates a `ServerOptions` instance from a file.

        If `file` is None, the default file location ('$VOLTTRON_HOME/config') is used.

        :param file: The path to the file containing the configuration options.
        :type file: Optional[Union[pathlib.Path, str]]

        :returns: A `_ServerOptions` instance created from the file.
        :rtype: _ServerOptions
        """
        if file is None:
            if os.environ.get('VOLTTRON_HOME'):
                file = Path(os.environ.get('VOLTTRON_HOME')).expanduser() / "config"
            else:
                file = Path("~/.volttron/config").expanduser()

        if isinstance(file, str):
            file = Path(file)

        if file.exists():
            parser = ConfigParser(strict=False)
            parser.read(file)

            kwargs = {}

            for field in fields(ServerOptions):
                try:
                    value = parser.get(section="volttron", option=field.name.replace('_', '-'))
                    if value == 'None':
                        value = None
                    elif value == 'False' or value == 'True':
                        value = eval(value)
                    elif field.name == 'service_config' or field.name == 'volttron_home':
                        value = Path(value)
                    elif field.name == 'address':
                        value = value.split('\n')
                    kwargs[field.name] = value
                except configparser.NoOptionError:
                    pass
            kwargs['initialized'] = True
            options = ServerOptions(**kwargs)
        else:
            options = ServerOptions()
            options.store(file)

        return options



import inspect
import logging
import warnings
from logging import Logger
import sys
import os

import volttron.utils.jsonapi as jsonapi


def get_default_client_log_config(level=logging.DEBUG) -> dict:
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {
                "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                "datefmt": "%Y-%m-%d %H:%M:%S"
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": level,
                "formatter": "simple",
                "stream": "ext://sys.stdout"
            },
            "file": {
                "class": "logging.FileHandler",
                "level": "INFO",
                "formatter": "simple",
                "filename": "myapp.log",
                "mode": "a"
            }
        },
        "loggers": {
            "volttron.client": {
                "level": level
            }
        },
        "root": {
            "level": level,
            "handlers": ["console"]
        }
    }


class JsonFormatter(logging.Formatter):

    def format(self, record):
        dct = record.__dict__.copy()
        dct["msg"] = record.getMessage()
        dct.pop("args")
        exc_info = dct.pop("exc_info", None)
        if exc_info:
            dct["exc_text"] = "".join(traceback.format_exception(*exc_info))
        return jsonapi.dumps(dct)


class AgentFormatter(logging.Formatter):

    def __init__(self, fmt=None, datefmt=None):
        if fmt is None:
            fmt = "%(asctime)s %(composite_name)s %(levelname)s: %(message)s"
        super(AgentFormatter, self).__init__(fmt=fmt, datefmt=datefmt)

    def composite_name(self, record):
        if record.name == "agents.log":
            cname = "(%(processName)s %(process)d) %(remote_name)s"
        elif record.name.startswith("agents.std"):
            cname = "(%(processName)s %(process)d) <{}>".format(record.name.split(".", 2)[1])
        else:
            cname = "() %(name)s"
        return cname % record.__dict__

    def format(self, record):
        if "composite_name" not in record.__dict__:
            record.__dict__["composite_name"] = self.composite_name(record)
        return super(AgentFormatter, self).format(record)


def setup_logging(level=logging.DEBUG, console=False):
    from volttron.utils.commands import isapipe

    root = logging.getLogger()
    if not root.handlers:
        handler = logging.StreamHandler()

        if isapipe(sys.stderr) and "_LAUNCHED_BY_PLATFORM" in os.environ:
            handler.setFormatter(JsonFormatter())
        elif console:
            # Below format is more readable for console
            handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
        else:
            fmt = "%(asctime)s %(name)s %(levelname)s: %(message)s"
            handler.setFormatter(logging.Formatter(fmt))
        if level != logging.DEBUG:
            # import it here so that when urllib3 imports the requests package, ssl would already got
            # monkey patched by gevent.
            # and this warning is needed only when log level is not debug
            from urllib3.exceptions import InsecureRequestWarning

            warnings.filterwarnings("ignore", category=InsecureRequestWarning)
        root.addHandler(handler)
    root.setLevel(level)

    logging.getLogger("volttron.messagebus").setLevel("INFO")
    logging.getLogger("volttron.server").setLevel("WARN")



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import inspect
import logging
import os
import sys
import traceback
import weakref

import gevent.local
from gevent.event import AsyncResult


from volttron.utils import jsonapi, jsonrpc
from volttron.client.known_identities import AUTH, CONTROL, CONTROL_CONNECTION
from ..decorators import annotate, annotations, dualmethod, spawn
from volttron.client.vip.agent.results import ResultsDictionary, counter
from volttron.client.vip.agent.subsystems.base import SubsystemBase
from volttron.client.vip.agent import VIPError

__all__ = ["RPC"]

_ROOT_PACKAGE_PATH = (os.path.dirname(__import__(__name__.split(".", 1)[0]).__path__[-1]) + os.sep)


_log = logging.getLogger(__name__)


def _isregex(obj):
    return (obj is not None and isinstance(obj, str) and len(obj) > 1 and obj[0] == obj[-1] == "/")


class Dispatcher(jsonrpc.Dispatcher):

    def __init__(self, methods, local):
        super(Dispatcher, self).__init__()
        self.methods = methods
        self.local = local
        self._results = ResultsDictionary()

    def serialize(self, json_obj):
        return jsonapi.dumps(json_obj)

    def deserialize(self, json_string):
        return jsonapi.loads(json_string)

    def batch_call(self, requests):
        methods = []
        results = []
        for notify, method, args, kwargs in requests:
            if notify:
                ident = None
            else:
                result = next(self._results)
                ident = result.ident
                results.append(result)
            methods.append((ident, method, args, kwargs))
        return super(Dispatcher, self).batch_call(methods), results

    def call(self, method, args=None, kwargs=None):
        # pylint: disable=arguments-differ
        result = next(self._results)
        return super(Dispatcher, self).call(result.ident, method, args, kwargs), result

    def result(self, response, ident, value, context=None):
        try:
            result = self._results.pop(ident)
        except KeyError:
            return
        result.set(value)

    def error(self, response, ident, code, message, data=None, context=None):
        try:
            result = self._results.pop(ident)
        except KeyError:
            return
        result.set_exception(jsonrpc.exception_from_json(code, message, data))

    def exception(self, response, ident, message, context=None):
        # XXX: Should probably wrap exception in RPC specific error
        #      rather than re-raising.
        (
            _,    # exc_type
            exc,
            _,    # exc_tb
        ) = sys.exc_info()    # pylint: disable=unused-variable
        try:
            result = self._results.pop(ident)
        except KeyError:
            return
        result.set_exception(exc)

    def method(self, request, ident, name, args, kwargs, batch=None, context=None):
        if kwargs:
            try:
                args, kwargs = kwargs["*args"], kwargs["**kwargs"]
            except KeyError:
                pass
        try:
            method = self.methods[name]
        except KeyError:
            if name == "inspect":
                return {"methods": list(self.methods)}
            elif name.endswith(".inspect"):
                try:
                    method = self.methods[name[:-8]]
                except KeyError:
                    pass
                else:
                    return self._inspect(method)
            raise NotImplementedError(name)
        local = self.local
        local.vip_message = context
        local.request = request
        local.batch = batch
        try:
            return method(*args, **kwargs)
        except Exception as exc:    # pylint: disable=broad-except
            exc_tb = traceback.format_exc()
            _log.error("unhandled exception in JSON-RPC method %r: \n%s", name, exc_tb)
            if getattr(method, "traceback", True):
                exc.exc_info = {"exc_tb": exc_tb}
            raise
        finally:
            del local.vip_message
            del local.request
            del local.batch

    @staticmethod
    def _inspect(method):
        response = {"params": {}}
        signature = inspect.signature(method)
        for p in signature.parameters.values():
            response["params"][p.name] = {"kind": p.kind.name}
            if p.default is not inspect.Parameter.empty:
                response["params"][p.name]["default"] = p.default
            if p.annotation is not inspect.Parameter.empty:
                annotation = (p.annotation.__name__ if type(p.annotation) is type else str(p.annotation))
                response["params"][p.name]["annotation"] = annotation
        doc = inspect.getdoc(method)
        if doc:
            response["doc"] = doc
        try:
            source = inspect.getsourcefile(method)
            cut = len(os.path.commonprefix([_ROOT_PACKAGE_PATH, source]))
            source = source[cut:]
            lineno = inspect.getsourcelines(method)[1]
        except Exception:
            pass
        else:
            response["source"] = {"file": source, "line_number": lineno}
        ret = signature.return_annotation
        if ret is not inspect.Signature.empty:
            response["return"] = \
                ret.__name__ if type(ret) is type else str(ret)
        return response


class RPC(SubsystemBase):

    def __init__(self, core, owner, peerlist_subsys):
        self.core = weakref.ref(core)
        self._owner = owner
        self.context = None
        self._exports = {}
        self._dispatcher = None
        self._counter = counter()
        self._outstanding = weakref.WeakValueDictionary()
        core.register("RPC", self._handle_subsystem, self._handle_error)
        core.register(
            "external_rpc",
            self._handle_external_rpc_subsystem,
            self._handle_error,
        )
        self._isconnected = True
        self.peerlist_subsystem = peerlist_subsys
        self.peer_list = {}
        self._protected_rpcs = None

        def export(member):    # pylint: disable=redefined-outer-name
            for name in annotations(member, set, "__rpc__.exports"):
                self._exports[name] = member

        inspect.getmembers(owner, export)

        def setup(sender, **kwargs):
            # pylint: disable=unused-argument
            self.context = gevent.local.local()
            self._dispatcher = Dispatcher(self._exports, self.context)
            self.export(self._add_protected_rpcs, "rpc.add_protected_rpcs")

        core.onsetup.connect(setup, self)
        core.ondisconnected.connect(self._disconnected)
        core.onconnected.connect(self._connected)

    def _connected(self, sender, **kwargs):
        self._isconnected = True
        # Registering to 'onadd' and 'ondrop' signals to get notified
        # whenever new peer is added/removed
        self.peerlist_subsystem.onadd.connect(self._add_new_peer)
        self.peerlist_subsystem.ondrop.connect(self._drop_new_peer)

    def _disconnected(self, sender, **kwargs):
        self._isconnected = False

    def _add_new_peer(self, sender, **kwargs):
        try:
            peer = kwargs.pop("peer")
            message_bus = kwargs.pop("message_bus")
            self.peer_list[peer] = message_bus
        except KeyError:
            pass

    def _drop_new_peer(self, sender, **kwargs):
        try:
            peer = kwargs.pop("peer")
            self.peer_list.pop(peer)
        except KeyError:
            pass

    def _add_auth_check(self, method):
        """
        Adds an authorization check to verify the calling agent has the
        required capabilities.
        """

        def checked_method(*args, **kwargs):
            calling_user = str(self.context.vip_message.user)
            #method_name = method.__name__
            # method.__name__ will give actual method name, but we want the alias used to export this method
            # ex: export("actual_method_name", "alias_exported_which_will_be_used_by_caller")
            method_name = self.context.vip_message.args[0]['method']
            #args_dict = inspect.getcallargs(method, *args, **kwargs)
            signature = inspect.signature(method)
            bound_args = signature.bind(*args, **kwargs)
            bound_args.apply_defaults()
            args_dict = bound_args.arguments
            # Remove self from args_dict if it exists to avoid sending the entire object across.
            # Fixes Issue https://github.com/eclipse-volttron/volttron-core/issues/198
            remove_self = args_dict.pop("self", None)
            from volttron.types.auth import AuthException
            try:
                self.call(AUTH,
                          method="check_rpc_authorization",
                          identity=calling_user,
                          method_name=f"{self.core().identity}.{method_name}",
                          method_args=args_dict).get(timeout=10)
            except AuthException as e:
                # msg = ("method '{}' requires capabilities {}, but capability {} "
                #        "was provided for user {}").format(method.__name__, required_caps,
                #                                           user_capabilites, user)
                raise jsonrpc.exception_from_json(jsonrpc.UNAUTHORIZED, e.args)
            return method(*args, **kwargs)

        return checked_method

    def _wrap_protected_rpcs(self, protected_rpcs):
        """
        Iterates over exported methods and adds authorization checks
        for protected methods
        """
        if protected_rpcs:
            for method_name in self._exports:
                # Don't wrap the method if it is already wrapped by checked_method
                # ideally we shouldn't be here.
                if method_name in protected_rpcs and self._exports[method_name].__name__ !="checked_method":
                    self._exports[method_name] = self._add_auth_check(self._exports[method_name])

    def _add_protected_rpcs(self, updated_list: list[str]):
        if not self._protected_rpcs:
            # nothing was there before so set variable and update all
            self._protected_rpcs = updated_list
            self._wrap_protected_rpcs(self._protected_rpcs)
        else:
            newly_protected = set(updated_list) - set(self._protected_rpcs)
            self._wrap_protected_rpcs(newly_protected)

    def _remove_protected_rpcs(self, remove_list: list[str]):
        if self._protected_rpcs is None:
            self._protected_rpcs = self.get_protected_rpcs()
        if self._protected_rpcs:
            for r in remove_list:
                if r in self._protected_rpcs:
                    method = getattr(self._owner, r, None)
                    # Verify the retrieved method object
                    if method and inspect.ismethod(method):
                        self._exports[r] = method
                    else:
                        raise ValueError(f"Method '{r}' not found in the instance or is not a method.")

    @spawn
    def _handle_external_rpc_subsystem(self, message):
        ret_msg = dict()
        operation = message.args[0]
        rpc_msg = message.args[1]    # jsonapi.loads(message.args[1])
        try:
            method_args = rpc_msg["args"]
            # message.args = [method_args]
            message.args = method_args
            for idx, msg in enumerate(message.args):
                if isinstance(msg, str):
                    message.args[idx] = jsonapi.loads(msg)
            dispatch = self._dispatcher.dispatch
            # _log.debug("External RPC IN message args {}".format(message))

            responses = [response for response in (dispatch(msg, message) for msg in message.args) if response]
            # _log.debug("External RPC Responses {}".format(responses))
            if responses:
                message.user = ""
                try:
                    message.peer = ""
                    message.subsystem = "external_rpc"
                    frames = []
                    operation = "send_platform"
                    frames.append(operation)
                    msg = jsonapi.dumps(
                        dict(
                            to_platform=rpc_msg["from_platform"],
                            to_peer=rpc_msg["from_peer"],
                            from_platform=rpc_msg["to_platform"],
                            from_peer=rpc_msg["to_peer"],
                            args=responses,
                        ))
                    frames.append(msg)
                except KeyError:
                    _log.error("External RPC message did not contain "
                               "proper message format")
                message.args = jsonapi.dumps(ret_msg)
                try:
                    self.core().connection.send_vip(
                        peer="",
                        subsystem="external_rpc",
                        args=frames,
                        msg_id=message.id,
                        user=message.user,
                        copy=False,
                    )
                except OSError as e:
                    # why are we only logging and not throwing the exception
                    # raise VIPError(e.errno, e.strerror, "", "external_rpc")
                    pass
        except KeyError:
            pass

    @spawn
    def _handle_subsystem(self, message):
        dispatch = self._dispatcher.dispatch

        responses = [response for response in (dispatch(msg, message) for msg in message.args) if response]
        if responses:
            message.user = ""
            message.args = responses
            try:
                if self._isconnected:
                    self.core().connection.send_vip_object(message, copy=False)
            except OSError as e:
                # why are we only logging
                pass
                #raise VIPError(e.errno, e.strerror, "", "RPC")

    def _handle_error(self, sender, message, error, **kwargs):
        result = self._outstanding.pop(message.id, None)
        if isinstance(result, AsyncResult):
            result.set_exception(error)
        elif result:
            for result in result:
                result.set_exception(error)

    def get_exports(self):
        """Returns a list copy of all exported methods."""
        return [method for method in self._exports].copy()

    @dualmethod
    def export(self, method, name=None):
        self._exports[name or method.__name__] = method
        return method

    @export.classmethod
    def export(cls, name=None):    # pylint: disable=no-self-argument
        if name is not None and not isinstance(name, str):
            method, name = name, name.__name__
            annotate(method, set, "__rpc__.exports", name)
            return method

        def decorate(method):
            annotate(method, set, "__rpc__.exports", name)
            return method

        return decorate

    def batch(self, peer, requests):
        request, results = self._dispatcher.batch_call(requests)
        if results:
            items = weakref.WeakSet(results)
            ident = "%s.%s" % (next(self._counter), id(items))
            for result in results:
                result._weak_set = items    # pylint: disable=protected-access
            self._outstanding[ident] = items
        else:
            ident = ""
        if request:
            if self._isconnected:
                try:
                    self.core().connection.send_vip(peer, "RPC", [request], msg_id=ident)
                except OSError as e:
                    pass
                    # we were only logging. why?
                    # raise VIPError(e.errno, e.strerror, peer, "RPC")
        return results or None

    def call(self, peer, method, *args, **kwargs):
        if self._protected_rpcs is None and peer not in [AUTH, CONTROL_CONNECTION, CONTROL]:
            # first rpc call
            # TODO: can this be done on onconnect? or on some other event
            self._protected_rpcs = self.get_protected_rpcs()
            self._wrap_protected_rpcs(self._protected_rpcs)
        self_ref = kwargs.pop("self", None)
        platform = kwargs.pop("external_platform", "")
        request, result = self._dispatcher.call(method, args, kwargs)
        ident = f"{next(self._counter)}.{hash(result)}"
        self._outstanding[ident] = result
        subsystem = None
        frames = []

        if not self._isconnected:
            return

        if platform == "":    # local platform
            subsystem = "RPC"
            frames.append(request)
        else:
            frames = []
            operation = "send_platform"
            subsystem = "external_rpc"
            frames.append(operation)
            msg = dict(
                to_platform=platform,
                to_peer=peer,
                from_platform="",
                from_peer="",
                args=[request],
            )
            frames.append(msg)
            peer = ""

        try:
            self.core().connection.send_vip(peer, subsystem, args=frames, msg_id=ident)
        except OSError as e:
            pass
            # we were only logging. why?
            # raise VIPError(e.errno, e.strerror, peer, "RPC")

        return result

    def get_protected_rpcs(self):
        return self.call(AUTH, "get_protected_rpcs", self._owner.core.identity).get(timeout=10)

    __call__ = call

    def notify(self, peer, method, *args, **kwargs):
        platform = kwargs.pop("external_platform", "")
        request = self._dispatcher.notify(method, args, kwargs)
        frames = []
        if not self._isconnected:
            return

        subsystem = None
        if platform == "":
            subsystem = "RPC"
            frames.append(request)
        else:
            operation = "send_platform"
            subsystem = "external_rpc"
            frames.append(operation)
            msg = dict(
                to_platform=platform,
                to_peer=peer,
                from_platform="",
                from_peer="",
                args=[request],
            )
            frames.append(msg)
            peer = ""

        try:
            self.core().connection.send_vip(peer, subsystem, args=frames)
        except OSError as e:
            pass
            # we were only logging. why?
            # raise VIPError(e.errno, e.strerror, peer, "RPC")
